"""Channels consumers for Resolwe Flow."""
import logging

from channels.consumer import SyncConsumer

from resolwe.flow.utils.purge import data_purge
from resolwe.utils import BraceMessage as __

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class PurgeConsumer(SyncConsumer):
    """Purge consumer."""

    def purge_run(self, event):
        """Run purge for the object with ``data_id`` specified in ``event`` argument."""
        data_id = event['data_id']
        verbosity = event['verbosity']

        try:
            logger.info(__("Running purge for Data id {}.", data_id))
            data_purge(data_ids=[data_id], verbosity=verbosity, delete=True)
        except Exception:  # pylint: disable=broad-except
            logger.exception("Error while purging data object.", extra={'data_id': data_id})



""".. Ignore pydocstyle D400.

================
Flow Serializers
================

Base Serializer
===============

Base Resolwe serializer.

.. autoclass:: resolwe.flow.serializers.base.ResolweBaseSerializer
    :members:

"""

from .base import NoContentError, ResolweBaseSerializer
from .collection import CollectionSerializer
from .contributor import ContributorSerializer
from .data import DataSerializer
from .descriptor import DescriptorSchemaSerializer
from .entity import EntitySerializer
from .process import ProcessSerializer
from .relation import RelationPartitionSerializer, RelationSerializer
from .storage import StorageSerializer



# pylint: disable=missing-docstring
from collections import OrderedDict

from guardian.shortcuts import assign_perm
from rest_framework.test import APIRequestFactory

from resolwe.flow.models import Collection, Data, DescriptorSchema, Entity, Process
from resolwe.flow.serializers import CollectionSerializer, DataSerializer, EntitySerializer
from resolwe.test import TestCase


class CollectionSerializerTest(TestCase):
    def setUp(self):
        super().setUp()

        process = Process.objects.create(
            name='Test process',
            contributor=self.contributor,
        )

        self.collection = Collection.objects.create(
            name='Test collection',
            contributor=self.contributor,
        )
        # In collection, user has `view` permission
        self.data_1 = Data.objects.create(
            name='First data',
            contributor=self.contributor,
            process=process,
        )
        # In collection, user has no permissions
        self.data_2 = Data.objects.create(
            name='Second data',
            contributor=self.contributor,
            process=process,
        )
        # Not in collection, user has `view` permission
        self.data_3 = Data.objects.create(
            name='Third data',
            contributor=self.contributor,
            process=process,
        )

        self.collection.data.add(self.data_1, self.data_2)

        assign_perm('view_collection', self.user, self.collection)
        assign_perm('view_data', self.user, self.data_1)
        assign_perm('view_data', self.user, self.data_3)

        self.factory = APIRequestFactory()

    def test_filter_data_permissions(self):
        request = self.factory.get('/')
        request.user = self.user
        request.query_params = {}

        serializer = CollectionSerializer(self.collection, context={'request': request})
        self.assertEqual(serializer.data['data'], [self.data_1.pk])


class EntitySerializerTest(TestCase):
    def setUp(self):
        super().setUp()

        process = Process.objects.create(
            name='Test process',
            contributor=self.contributor,
        )

        self.entity = Entity.objects.create(
            name='Test entity',
            contributor=self.contributor,
        )
        # In entity, user has `view` permission
        self.data_1 = Data.objects.create(
            name='First data',
            contributor=self.contributor,
            process=process,
        )
        # In entity, user has no permissions
        self.data_2 = Data.objects.create(
            name='Second data',
            contributor=self.contributor,
            process=process,
        )
        # Not in entity, user has `view` permission
        self.data_3 = Data.objects.create(
            name='Third data',
            contributor=self.contributor,
            process=process,
        )

        self.entity.data.add(self.data_1, self.data_2)

        assign_perm('view_entity', self.user, self.entity)
        assign_perm('view_data', self.user, self.data_1)
        assign_perm('view_data', self.user, self.data_3)

        self.factory = APIRequestFactory()

    def test_filter_data_permissions(self):
        request = self.factory.get('/')
        request.user = self.user
        request.query_params = {}

        serializer = EntitySerializer(self.entity, context={'request': request})
        self.assertEqual(serializer.data['data'], [self.data_1.pk])


class ResolweSlugRelatedFieldTest(TestCase):
    def setUp(self):
        super().setUp()

        self.process = Process.objects.create(
            slug='test-process',
            contributor=self.contributor,
        )
        assign_perm('view_process', self.user, self.process)

        self.descriptor_schema1 = DescriptorSchema.objects.create(
            slug='test-schema',
            contributor=self.contributor,
            version='1.0.0',
        )
        assign_perm('view_descriptorschema', self.user, self.descriptor_schema1)

        self.descriptor_schema2 = DescriptorSchema.objects.create(
            slug='test-schema',
            contributor=self.contributor,
            version='2.0.0',
        )
        assign_perm('view_descriptorschema', self.user, self.descriptor_schema2)

        self.descriptor_schema3 = DescriptorSchema.objects.create(
            slug='test-schema',
            contributor=self.contributor,
            version='3.0.0',
        )

        self.factory = APIRequestFactory()

    def test_to_internal_value(self):
        request = self.factory.get('/')
        request.user = self.user
        request.query_params = {}
        data = {
            'contributor': self.user.pk,
            'process': 'test-process',
            'descriptor_schema': 'test-schema',
        }

        serializer = DataSerializer(data=data, context={'request': request})
        # is_valid() needs to be called before accessing ``validated_data``
        serializer.is_valid()
        # Check that descriptor schmena with highest version & view permission is used:
        self.assertEqual(serializer.validated_data['descriptor_schema'], self.descriptor_schema2)

    def test_to_representation(self):
        request = self.factory.get('/')
        request.user = self.user
        request.query_params = {}

        data = Data.objects.create(
            contributor=self.user,
            process=self.process,
            descriptor_schema=self.descriptor_schema1,
        )

        serializer = DataSerializer(data, context={'request': request})
        self.assertEqual(serializer.data['process'], self.process.pk)

        # Check that descriptor_schema is properly hydrated (but remove
        # values that are not deterministic from the checking procedure)
        descriptor_schema_hydrated = serializer.data['descriptor_schema']
        for key in ['created', 'modified', 'id']:
            self.assertTrue(key in descriptor_schema_hydrated)
            descriptor_schema_hydrated.pop(key)
        descriptor_schema_hydrated.get('contributor', {}).pop('id')
        self.assertDictEqual(descriptor_schema_hydrated, {
            'slug': 'test-schema',
            'version': '1.0.0',
            'name': '',
            'description': '',
            'schema': [],
            'contributor': OrderedDict([
                ('first_name', 'Joe'),
                ('last_name', 'Miller'),
                ('username', 'contributor'),
            ]),
        })



# pylint: disable=missing-docstring
import datetime

from django.contrib.auth import get_user_model

from guardian.shortcuts import assign_perm
from rest_framework import status
from rest_framework.test import APIRequestFactory, force_authenticate

from resolwe.flow.filters import CollectionFilter, DataFilter, EntityFilter, ProcessFilter
from resolwe.flow.models import Collection, Data, DataDependency, DescriptorSchema, Entity, Process
from resolwe.flow.views import CollectionViewSet, DataViewSet
from resolwe.test import TestCase

factory = APIRequestFactory()  # pylint: disable=invalid-name


class DataFilterTestCase(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user_1 = get_user_model().objects.create(username="first_user")
        cls.user_2 = get_user_model().objects.create(username="second_user")

        cls.proc_1 = Process.objects.create(
            name='Test process 1',
            slug='test-process-1',
            contributor=cls.user_1,
            type='data:test:'
        )

        cls.proc_2 = Process.objects.create(
            name='Test process 2',
            slug='test-process-2',
            contributor=cls.user_1,
            type='data:test:second:'
        )

        cls.proc_3 = Process.objects.create(
            name='Test process 3',
            slug='test-process-3',
            contributor=cls.user_1,
            type='data:third:'
        )

        cls.data_1 = Data.objects.create(
            name='Test data 1',
            slug='test-data-1',
            contributor=cls.user_1,
            process=cls.proc_1,
            status=Data.STATUS_DONE,
            started=datetime.datetime(2016, 7, 30, 14, 0),
            finished=datetime.datetime(2016, 7, 30, 14, 30),
            size=42,
        )
        cls.data_1.created = datetime.datetime(2016, 7, 30, 13, 59)
        cls.data_1.save()

        cls.data_2 = Data.objects.create(
            name='Test data 2',
            slug='test-data-2',
            contributor=cls.user_1,
            process=cls.proc_2,
            status=Data.STATUS_ERROR,
            started=datetime.datetime(2016, 8, 30, 15, 0),
            finished=datetime.datetime(2016, 8, 30, 15, 30),
            tags=['foo', 'bar', 'moo'],
            size=42,
        )
        cls.data_2.created = datetime.datetime(2016, 8, 30, 14, 59)
        cls.data_2.save()

        cls.data_3 = Data.objects.create(
            name='Another data object',
            slug='another-data-object',
            contributor=cls.user_2,
            process=cls.proc_3,
            status=Data.STATUS_DONE,
            started=datetime.datetime(2013, 1, 15, 8, 0),
            finished=datetime.datetime(2013, 1, 15, 10, 0),
            tags=['bar']
        )
        cls.data_3.created = datetime.datetime(2013, 1, 15, 7, 59)
        cls.data_3.save()

        cls.collection = Collection.objects.create(
            name='Test collection',
            slug='test-collection',
            contributor=cls.user_1
        )
        cls.collection.data.add(cls.data_1)

    def _apply_filter(self, filters, expected):
        filtered = DataFilter(filters, queryset=Data.objects.all())
        self.assertCountEqual(filtered.qs, expected)

    def test_filter_id(self):
        self._apply_filter({'id': self.data_1.pk}, [self.data_1])
        self._apply_filter({'id__in': '{},{}'.format(self.data_1.pk, self.data_2.pk)},
                           [self.data_1, self.data_2])

    def test_filter_slug(self):
        self._apply_filter({'slug': 'test-data-2'}, [self.data_2])
        self._apply_filter({'slug__startswith': 'test'}, [self.data_1, self.data_2])

    def test_filter_name(self):
        self._apply_filter({'name': 'Test data 2'}, [self.data_2])
        self._apply_filter({'name__startswith': 'Test'}, [self.data_1, self.data_2])

    def test_filter_contributor(self):
        self._apply_filter({'contributor': self.user_1.pk}, [self.data_1, self.data_2])

    def test_filter_created(self):
        self._apply_filter({'created__year': '2016'}, [self.data_1, self.data_2])
        self._apply_filter({'created__year__gt': '2015'}, [self.data_1, self.data_2])
        self._apply_filter({'created__year__gt': '2015', 'created__month__gte': '8'}, [self.data_2])

    def test_filter_modified(self):
        year = datetime.date.today().strftime('%Y')
        self._apply_filter({'modified__year': year}, [self.data_1, self.data_2, self.data_3])
        self._apply_filter({'modified__year__gt': year}, [])

    def test_filter_collection(self):
        self._apply_filter({'collection': self.collection.pk}, [self.data_1])
        self._apply_filter({'collection__slug': 'test-collection'}, [self.data_1])

    def test_filter_type(self):
        self._apply_filter({'type': 'data:test:'}, [self.data_1, self.data_2])

    def test_filter_status(self):
        self._apply_filter({'status': 'ok'}, [self.data_1, self.data_3])

    def test_filter_finished(self):
        self._apply_filter({'finished__year': '2016'}, [self.data_1, self.data_2])
        self._apply_filter({'finished__year__gt': '2015'}, [self.data_1, self.data_2])
        self._apply_filter({'finished__year__gt': '2015', 'finished__month__gte': '8'}, [self.data_2])

    def test_filter_started(self):
        self._apply_filter({'started__year': '2016'}, [self.data_1, self.data_2])
        self._apply_filter({'started__year__gt': '2015'}, [self.data_1, self.data_2])
        self._apply_filter({'started__year__gt': '2015', 'started__month__gte': '8'}, [self.data_2])

    def test_filter_process(self):
        self._apply_filter({'process': self.proc_1.pk}, [self.data_1])
        self._apply_filter({'process__slug': 'test-process-1'}, [self.data_1])

    def test_filter_tags(self):
        self._apply_filter({'tags': 'foo'}, [self.data_2])
        self._apply_filter({'tags': 'bar'}, [self.data_2, self.data_3])
        self._apply_filter({'tags': 'bar,moo'}, [self.data_2])

    def test_filter_parents_children(self):
        self._apply_filter({'parents': self.data_1.pk}, [])
        self._apply_filter({'parents': self.data_2.pk}, [])
        self._apply_filter({'children': self.data_1.pk}, [])
        self._apply_filter({'children': self.data_2.pk}, [])

        DataDependency.objects.create(
            parent=self.data_1,
            child=self.data_2,
            kind=DataDependency.KIND_IO,
        )

        self._apply_filter({'parents': self.data_1.pk}, Data.objects.filter(parents=self.data_1))
        self._apply_filter({'parents': self.data_2.pk}, Data.objects.filter(parents=self.data_2))
        self._apply_filter({'children': self.data_1.pk}, Data.objects.filter(children=self.data_1))
        self._apply_filter({'children': self.data_2.pk}, Data.objects.filter(children=self.data_2))


class TestDataViewSetFilters(TestCase):

    def setUp(self):
        super().setUp()

        self.data_viewset = DataViewSet.as_view(actions={
            'get': 'list',
        })

        self.collection = Collection.objects.create(contributor=self.contributor)

        self.proc1 = Process.objects.create(
            type='data:test:process1:',
            name='First process',
            slug='test-process-1',
            version='1.0.0',
            contributor=self.contributor,
            entity_type='test-schema',
            entity_descriptor_schema='test-schema',
            input_schema=[{'name': 'input_data', 'type': 'data:test:', 'required': False}],
        )

        self.proc2 = Process.objects.create(
            type='data:test:process2:',
            name='Second process',
            slug='test-process-2',
            version='1.0.0',
            contributor=self.contributor,
            entity_type='test-schema',
            entity_descriptor_schema='test-schema',
            input_schema=[{'name': 'input_data', 'type': 'data:test:', 'required': False}],
        )

        self.descriptor_schema = DescriptorSchema.objects.create(
            slug='test-schema',
            version='1.0.0',
            contributor=self.contributor,
        )

        self.data = []
        for index in range(10):
            data = Data.objects.create(
                name='Data {}'.format(index),
                slug='dataslug-{}'.format(index),
                contributor=self.contributor,
                process=self.proc1 if index < 5 else self.proc2,
                status=Data.STATUS_DONE if index > 0 else Data.STATUS_RESOLVING,
                started=datetime.datetime(2016, 7, 31, index, 0),
                finished=datetime.datetime(2016, 7, 31, index, 30),
                tags=['foo', 'index{}'.format(index)],
            )

            data.created = datetime.datetime(2016, 7, 30, index, 59)
            data.save()

            assign_perm('owner_data', self.admin, data)

            if index == 0:
                self.collection.data.add(data)

            self.data.append(data)

    def _check_filter(self, query_args, expected):
        request = factory.get('/', query_args, format='json')
        force_authenticate(request, self.admin)
        response = self.data_viewset(request)
        expected = [item.pk for item in expected]

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data), len(expected))
        for item in response.data:
            self.assertIn(item['id'], expected)

    def test_filter_id(self):
        self._check_filter({'id': str(self.data[0].pk)}, [self.data[0]])
        self._check_filter({'id': str(self.data[2].pk)}, [self.data[2]])
        self._check_filter({'id__in': '{},{}'.format(self.data[0].pk, self.data[2].pk)}, [self.data[0], self.data[2]])

    def test_filter_slug(self):
        self._check_filter({'slug': 'dataslug-1'}, [self.data[1]])
        self._check_filter({'slug': 'dataslug-5'}, [self.data[5]])
        self._check_filter({'slug__in': 'dataslug-1'}, [self.data[1]])
        self._check_filter({'slug__in': 'dataslug-1,dataslug-5'}, [self.data[1], self.data[5]])

    def test_filter_name(self):
        self._check_filter({'name': 'Data 1'}, [self.data[1]])
        self._check_filter({'name': 'data 1'}, [])
        self._check_filter({'name': 'Data 2'}, [self.data[2]])
        self._check_filter({'name': 'Data'}, [])
        self._check_filter({'name': '1'}, [])

    def test_filter_contributor(self):
        self._check_filter({'contributor': str(self.contributor.pk)}, self.data)
        self._check_filter({'contributor': '0'}, [])
        self._check_filter({'contributor__in': '0,{}'.format(self.contributor.pk)}, self.data)

    def test_filter_owners(self):
        self._check_filter({'owners': str(self.admin.pk)}, self.data)
        self._check_filter({'owners': '0'}, [])
        self._check_filter({'owners__in': '0,{}'.format(self.admin.pk)}, self.data)

    def test_filter_created(self):
        self._check_filter({'created': self.data[0].created.isoformat()}, [self.data[0]])
        self._check_filter({'created__gt': '2016'}, self.data)
        self._check_filter({'created__gt': '2016-07-30T05:59:00'}, self.data[6:])
        self._check_filter({'created__gte': '2016-07-30T05:59:00'}, self.data[5:])
        self._check_filter({'created__lt': '2016'}, [])
        self._check_filter({'created__lt': '2016-07-30T05:59:00'}, self.data[:5])
        self._check_filter({'created__lte': '2016-07-30T05:59:00'}, self.data[:6])

    def test_filter_modified(self):
        now = self.data[0].modified
        five_offset = self.data[5].modified
        self._check_filter({'modified': now.isoformat()}, [self.data[0]])
        self._check_filter({'modified__gt': str(now.year)}, self.data)
        self._check_filter({'modified__gt': five_offset.isoformat()}, self.data[6:])
        self._check_filter({'modified__gte': five_offset.isoformat()}, self.data[5:])
        self._check_filter({'modified__lt': str(now.year)}, [])
        self._check_filter({'modified__lt': five_offset.isoformat()}, self.data[:5])
        self._check_filter({'modified__lte': five_offset.isoformat()}, self.data[:6])

    def test_filter_started(self):
        self._check_filter({'started': self.data[0].started.isoformat()}, [self.data[0]])
        self._check_filter({'started__gt': '2016'}, self.data)
        self._check_filter({'started__gt': '2016-07-31T05:00:00'}, self.data[6:])
        self._check_filter({'started__gte': '2016-07-31T05:00:00'}, self.data[5:])
        self._check_filter({'started__lt': '2016'}, [])
        self._check_filter({'started__lt': '2016-07-31T05:00:00'}, self.data[:5])
        self._check_filter({'started__lte': '2016-07-31T05:00:00'}, self.data[:6])

    def test_filter_finished(self):
        self._check_filter({'finished': self.data[0].finished.isoformat()}, [self.data[0]])
        self._check_filter({'finished__gt': '2016'}, self.data)
        self._check_filter({'finished__gt': '2016-07-31T05:30:00'}, self.data[6:])
        self._check_filter({'finished__gte': '2016-07-31T05:30:00'}, self.data[5:])
        self._check_filter({'finished__lt': '2016'}, [])
        self._check_filter({'finished__lt': '2016-07-31T05:30:00'}, self.data[:5])
        self._check_filter({'finished__lte': '2016-07-31T05:30:00'}, self.data[:6])

    def test_filter_collection(self):
        self._check_filter({'collection': str(self.collection.pk)}, [self.data[0]])

    def test_filter_type(self):
        self._check_filter({'type': 'data'}, self.data)
        self._check_filter({'type': 'data:test'}, self.data)
        self._check_filter({'type': 'data:test:process1'}, self.data[:5])
        self._check_filter({'type': 'data:test:process2'}, self.data[5:])
        self._check_filter({'type': 'data:'}, [])
        self._check_filter({'type': 'data:test:'}, [])
        self._check_filter({'type': 'data:test:process1:'}, self.data[:5])
        self._check_filter({'type': 'data:test:process2:'}, self.data[5:])
        self._check_filter({'type__exact': 'data:test'}, [])
        self._check_filter({'type__exact': 'data:test:process1:'}, self.data[:5])
        self._check_filter({'type__exact': 'data:test:process2:'}, self.data[5:])

    def test_filter_status(self):
        self._check_filter({'status': 'OK'}, self.data[1:])
        self._check_filter({'status': 'RE'}, [self.data[0]])
        self._check_filter({'status__in': 'OK,RE'}, self.data)

    def test_filter_process(self):
        self._check_filter({'process': str(self.proc1.pk)}, self.data[:5])
        self._check_filter({'process': str(self.proc2.pk)}, self.data[5:])

    def test_filter_process_name(self):
        self._check_filter({'process_name': 'first'}, self.data[:5])
        self._check_filter({'process_name': 'fir'}, self.data[:5])
        self._check_filter({'process_name': 'rst'}, [])
        self._check_filter({'process_name': 'second'}, self.data[5:])
        self._check_filter({'process_name': 'sec'}, self.data[5:])

    def test_filter_tags(self):
        self._check_filter({'tags': 'foo'}, self.data)
        self._check_filter({'tags': 'foo,index1'}, [self.data[1]])
        self._check_filter({'tags': 'foo,index5'}, [self.data[5]])
        self._check_filter({'tags': 'foo,index10'}, [])
        self._check_filter({'tags': 'bar'}, [])

    def test_filter_parents_children(self):
        self._check_filter({'parents': self.data[0].pk}, [])
        self._check_filter({'parents': self.data[1].pk}, [])
        self._check_filter({'children': self.data[0].pk}, [])
        self._check_filter({'children': self.data[1].pk}, [])

        dependency = DataDependency.objects.create(
            parent=self.data[0],
            child=self.data[1],
            kind=DataDependency.KIND_IO,
        )

        self._check_filter({'parents': self.data[0].pk}, Data.objects.filter(parents=self.data[0]))
        self._check_filter({'parents': self.data[1].pk}, Data.objects.filter(parents=self.data[1]))
        self._check_filter({'children': self.data[0].pk}, Data.objects.filter(children=self.data[0]))
        self._check_filter({'children': self.data[1].pk}, Data.objects.filter(children=self.data[1]))

        dependency.delete()

        self._check_filter({'parents': self.data[0].pk}, [])
        self._check_filter({'parents': self.data[1].pk}, [])
        self._check_filter({'children': self.data[0].pk}, [])
        self._check_filter({'children': self.data[1].pk}, [])

    def test_filter_text(self):
        # By slug.
        self._check_filter({'text': 'dataslug-1'}, [self.data[1]])
        self._check_filter({'text': 'datasl'}, self.data)

        # By name.
        self._check_filter({'text': 'Data 1'}, [self.data[1]])
        self._check_filter({'text': 'data 2'}, [self.data[2]])
        self._check_filter({'text': 'Data'}, self.data)
        self._check_filter({'text': 'data'}, self.data)
        self._check_filter({'text': 'dat'}, self.data)
        self._check_filter({'text': 'ata'}, [])

        # By contributor.
        self._check_filter({'text': 'joe'}, self.data)
        self._check_filter({'text': 'oe'}, [])
        self._check_filter({'text': 'Miller'}, self.data)
        self._check_filter({'text': 'mill'}, self.data)

        # By owner.
        self._check_filter({'text': 'james'}, self.data)
        self._check_filter({'text': 'mes'}, [])
        self._check_filter({'text': 'Smith'}, self.data)
        self._check_filter({'text': 'smi'}, self.data)

        # By process name.
        self._check_filter({'text': 'first'}, self.data[:5])
        self._check_filter({'text': 'fir'}, self.data[:5])
        self._check_filter({'text': 'rst'}, [])
        self._check_filter({'text': 'process'}, self.data)


class CollectionFilterTestCase(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user_1 = get_user_model().objects.create(username="first_user")
        cls.user_2 = get_user_model().objects.create(username="second_user")

        cls.descriptor_schema_1 = DescriptorSchema.objects.create(
            contributor=cls.user_1,
            name='Descriptor schema 1'
        )

        cls.descriptor_schema_2 = DescriptorSchema.objects.create(
            contributor=cls.user_2,
            name='Descriptor schema 2'
        )

        cls.collection_1 = Collection.objects.create(
            name='Test collection 1',
            slug='test-collection-1',
            descriptor_schema=cls.descriptor_schema_1,
            contributor=cls.user_1
        )
        cls.collection_1.created = datetime.datetime(2016, 7, 30, 13, 59)
        cls.collection_1.save()

        cls.collection_2 = Collection.objects.create(
            name='Test collection 2',
            slug='test-collection-2',
            descriptor_schema=cls.descriptor_schema_1,
            contributor=cls.user_1
        )
        cls.collection_2.created = datetime.datetime(2016, 8, 30, 14, 59)
        cls.collection_2.save()

        cls.collection_3 = Collection.objects.create(
            name='Another collection',
            slug='another-collection',
            descriptor_schema=cls.descriptor_schema_2,
            contributor=cls.user_2
        )
        cls.collection_3.created = datetime.datetime(2013, 1, 15, 7, 59)
        cls.collection_3.save()

        cls.proc = Process.objects.create(
            name='Test process',
            slug='test-process',
            contributor=cls.user_1,
            type='data:test:'
        )

        cls.data = Data.objects.create(
            name='Test data',
            slug='test-data',
            contributor=cls.user_1,
            process=cls.proc,
            status=Data.STATUS_DONE
        )
        cls.collection_1.data.add(cls.data)
        cls.collection_2.data.add(cls.data)

    def _apply_filter(self, filters, expected):
        filtered = CollectionFilter(filters, queryset=Collection.objects.all())
        self.assertCountEqual(filtered.qs, expected)

    def test_filter_id(self):
        self._apply_filter({'id': self.collection_1.pk}, [self.collection_1])
        self._apply_filter({'id__in': '{},{}'.format(self.collection_1.pk, self.collection_2.pk)},
                           [self.collection_1, self.collection_2])

    def test_filter_slug(self):
        self._apply_filter({'slug': 'test-collection-2'}, [self.collection_2])
        self._apply_filter({'slug__startswith': 'test'}, [self.collection_1, self.collection_2])

    def test_filter_name(self):
        self._apply_filter({'name': 'Test collection 2'}, [self.collection_2])
        self._apply_filter({'name__startswith': 'Test'}, [self.collection_1, self.collection_2])

    def test_filter_contributor(self):
        self._apply_filter({'contributor': self.user_1.pk},
                           [self.collection_1, self.collection_2])

    def test_filter_created(self):
        self._apply_filter({'created__year': '2016'}, [self.collection_1, self.collection_2])
        self._apply_filter({'created__year__gt': '2015'}, [self.collection_1, self.collection_2])
        self._apply_filter({'created__year__gt': '2015', 'created__month__gte': '8'}, [self.collection_2])

    def test_filter_modified(self):
        year = datetime.date.today().strftime('%Y')
        self._apply_filter({'modified__year': year}, [self.collection_1, self.collection_2, self.collection_3])
        self._apply_filter({'modified__year__gt': year}, [])

    def test_filter_data(self):
        self._apply_filter({'data': self.data.pk}, [self.collection_1, self.collection_2])

    def test_filter_descriptor_schema(self):
        self._apply_filter({'descriptor_schema': self.descriptor_schema_1.pk}, [self.collection_1, self.collection_2])


class TestCollectionViewSetFilters(TestCase):

    def setUp(self):
        super().setUp()

        self.collection_viewset = CollectionViewSet.as_view(actions={
            'get': 'list',
        })

        self.collection = Collection.objects.create(
            name='Test collection 1',
            contributor=self.contributor,
        )

    def _check_filter(self, query_args, expected):
        request = factory.get('/', query_args, format='json')
        force_authenticate(request, self.admin)
        response = self.collection_viewset(request)
        expected = [item.pk for item in expected]

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data), len(expected))
        for item in response.data:
            self.assertIn(item['id'], expected)

    def test_filter_name(self):
        self._check_filter({'name': 'Test collection 1'}, [self.collection])
        self._check_filter({'name': 'Test collection'}, [])


class EntityFilterTestCase(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user_1 = get_user_model().objects.create(username="first_user")

        cls.entity_1 = Entity.objects.create(
            contributor=cls.user_1,
            descriptor_completed=True,
        )

        cls.entity_2 = Entity.objects.create(
            contributor=cls.user_1,
            descriptor_completed=False,
        )

    def _apply_filter(self, filters, expected):
        filtered = EntityFilter(filters, queryset=Entity.objects.all())
        self.assertCountEqual(filtered.qs, expected)

    def test_descriptor_completed(self):
        self._apply_filter({'descriptor_completed': 'true'}, [self.entity_1])
        self._apply_filter({'descriptor_completed': '1'}, [self.entity_1])
        self._apply_filter({'descriptor_completed': 'false'}, [self.entity_2])
        self._apply_filter({'descriptor_completed': '0'}, [self.entity_2])


class ProcessFilterTestCase(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = get_user_model().objects.create(username="first_user")

        cls.proc_1 = Process.objects.create(
            contributor=cls.user,
            type='data:alignment:bam:',
            category='analyses:alignment:',
            scheduling_class='BA',
        )

        cls.proc_2 = Process.objects.create(
            contributor=cls.user,
            type='data:expression:',
            category='analyses:',
            scheduling_class='IN',
        )

    def _apply_filter(self, filters, expected):
        filtered = ProcessFilter(filters, queryset=Process.objects.all())
        self.assertCountEqual(filtered.qs, expected)

    def test_category(self):
        self._apply_filter({'category': 'analyses:'}, [self.proc_1, self.proc_2])
        self._apply_filter({'category': 'analyses:alignment'}, [self.proc_1])

    def test_type(self):
        self._apply_filter({'type': 'data:'}, [self.proc_1, self.proc_2])
        self._apply_filter({'type': 'data:alignment:bam'}, [self.proc_1])
        self._apply_filter({'type': 'data:expression'}, [self.proc_2])

    def test_scheduling_class(self):
        self._apply_filter({'scheduling_class': 'BA'}, [self.proc_1])
        self._apply_filter({'scheduling_class': 'IN'}, [self.proc_2])

    def test_is_active(self):
        self._apply_filter({'is_active': True}, [self.proc_1, self.proc_2])
        self._apply_filter({'is_active': False}, [])

        self.proc_1.is_active = False
        self.proc_1.save()

        self._apply_filter({'is_active': True}, [self.proc_2])
        self._apply_filter({'is_active': False}, [self.proc_1])



""".. Ignore pydocstyle D400.

.. autoclass:: resolwe.flow.executors.docker.run.FlowExecutor
    :members:

"""
# pylint: disable=logging-format-interpolation
import asyncio
import json
import logging
import os
import shlex
import tempfile
import time
from asyncio import subprocess

from ..global_settings import PROCESS_META, SETTINGS
from ..local.run import FlowExecutor as LocalFlowExecutor
from ..protocol import ExecutorFiles
from . import constants
from .seccomp import SECCOMP_POLICY

DOCKER_START_TIMEOUT = 60

# Limits of containers' access to memory. We set the limit to ensure
# processes are stable and do not get killed by OOM signal.
DOCKER_MEMORY_HARD_LIMIT_BUFFER = 100
DOCKER_MEMORY_SWAP_RATIO = 2
DOCKER_MEMORY_SWAPPINESS = 1

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class FlowExecutor(LocalFlowExecutor):
    """Docker executor."""

    name = 'docker'

    def __init__(self, *args, **kwargs):
        """Initialize attributes."""
        super().__init__(*args, **kwargs)

        self.container_name_prefix = None
        self.tools_volumes = None
        self.temporary_files = []
        self.command = SETTINGS.get('FLOW_DOCKER_COMMAND', 'docker')

    def _generate_container_name(self):
        """Generate unique container name."""
        return '{}_{}'.format(self.container_name_prefix, self.data_id)

    async def start(self):
        """Start process execution."""
        # arguments passed to the Docker command
        command_args = {
            'command': self.command,
            'container_image': self.requirements.get('image', constants.DEFAULT_CONTAINER_IMAGE),
        }

        # Get limit defaults.
        limit_defaults = SETTINGS.get('FLOW_PROCESS_RESOURCE_DEFAULTS', {})

        # Set resource limits.
        limits = []
        # Each core is equivalent to 1024 CPU shares. The default for Docker containers
        # is 1024 shares (we don't need to explicitly set that).
        limits.append('--cpu-shares={}'.format(int(self.process['resource_limits']['cores']) * 1024))

        # Some SWAP is needed to avoid OOM signal. Swappiness is low to prevent
        # extensive usage of SWAP (this would reduce the performance).
        memory = self.process['resource_limits']['memory'] + DOCKER_MEMORY_HARD_LIMIT_BUFFER
        memory_swap = int(memory * DOCKER_MEMORY_SWAP_RATIO)

        limits.append('--memory={}m'.format(memory))
        limits.append('--memory-swap={}m'.format(memory_swap))
        limits.append('--memory-reservation={}m'.format(self.process['resource_limits']['memory']))
        limits.append('--memory-swappiness={}'.format(DOCKER_MEMORY_SWAPPINESS))

        # Set ulimits for interactive processes to prevent them from running too long.
        if self.process['scheduling_class'] == PROCESS_META['SCHEDULING_CLASS_INTERACTIVE']:
            # TODO: This is not very good as each child gets the same limit.
            limits.append('--ulimit cpu={}'.format(limit_defaults.get('cpu_time_interactive', 30)))

        command_args['limits'] = ' '.join(limits)

        # set container name
        self.container_name_prefix = SETTINGS.get('FLOW_EXECUTOR', {}).get('CONTAINER_NAME_PREFIX', 'resolwe')
        command_args['container_name'] = '--name={}'.format(self._generate_container_name())

        if 'network' in self.resources:
            # Configure Docker network mode for the container (if specified).
            # By default, current Docker versions use the 'bridge' mode which
            # creates a network stack on the default Docker bridge.
            network = SETTINGS.get('FLOW_EXECUTOR', {}).get('NETWORK', '')
            command_args['network'] = '--net={}'.format(network) if network else ''
        else:
            # No network if not specified.
            command_args['network'] = '--net=none'

        # Security options.
        security = []

        # Generate and set seccomp policy to limit syscalls.
        policy_file = tempfile.NamedTemporaryFile(mode='w')
        json.dump(SECCOMP_POLICY, policy_file)
        policy_file.file.flush()
        if not SETTINGS.get('FLOW_DOCKER_DISABLE_SECCOMP', False):
            security.append('--security-opt seccomp={}'.format(policy_file.name))
        self.temporary_files.append(policy_file)

        # Drop all capabilities and only add ones that are needed.
        security.append('--cap-drop=all')

        command_args['security'] = ' '.join(security)

        # Setup Docker volumes.
        def new_volume(kind, base_dir_name, volume, path=None, read_only=True):
            """Generate a new volume entry.

            :param kind: Kind of volume, which is used for getting extra options from
                settings (the ``FLOW_DOCKER_VOLUME_EXTRA_OPTIONS`` setting)
            :param base_dir_name: Name of base directory setting for volume source path
            :param volume: Destination volume mount point
            :param path: Optional additional path atoms appended to source path
            :param read_only: True to make the volume read-only
            """
            if path is None:
                path = []

            path = [str(atom) for atom in path]

            options = set(SETTINGS.get('FLOW_DOCKER_VOLUME_EXTRA_OPTIONS', {}).get(kind, '').split(','))
            options.discard('')
            # Do not allow modification of read-only option.
            options.discard('ro')
            options.discard('rw')

            if read_only:
                options.add('ro')
            else:
                options.add('rw')

            return {
                'src': os.path.join(SETTINGS['FLOW_EXECUTOR'].get(base_dir_name, ''), *path),
                'dest': volume,
                'options': ','.join(options),
            }

        volumes = [
            new_volume('data', 'DATA_DIR', constants.DATA_VOLUME, [self.data_id], read_only=False),
            new_volume('data_all', 'DATA_DIR', constants.DATA_ALL_VOLUME),
            new_volume('upload', 'UPLOAD_DIR', constants.UPLOAD_VOLUME, read_only=False),
            new_volume('secrets', 'RUNTIME_DIR', constants.SECRETS_VOLUME, [self.data_id, ExecutorFiles.SECRETS_DIR]),
        ]

        # Generate dummy passwd and create mappings for it. This is required because some tools
        # inside the container may try to lookup the given UID/GID and will crash if they don't
        # exist. So we create minimal user/group files.
        passwd_file = tempfile.NamedTemporaryFile(mode='w')
        passwd_file.write('root:x:0:0:root:/root:/bin/bash\n')
        passwd_file.write('user:x:{}:{}:user:/:/bin/bash\n'.format(os.getuid(), os.getgid()))
        passwd_file.file.flush()
        self.temporary_files.append(passwd_file)

        group_file = tempfile.NamedTemporaryFile(mode='w')
        group_file.write('root:x:0:\n')
        group_file.write('user:x:{}:user\n'.format(os.getgid()))
        group_file.file.flush()
        self.temporary_files.append(group_file)

        volumes += [
            new_volume('users', None, '/etc/passwd', [passwd_file.name]),
            new_volume('users', None, '/etc/group', [group_file.name]),
        ]

        # Create volumes for tools.
        # NOTE: To prevent processes tampering with tools, all tools are mounted read-only
        self.tools_volumes = []
        for index, tool in enumerate(self.get_tools_paths()):
            self.tools_volumes.append(new_volume(
                'tools',
                None,
                os.path.join('/usr/local/bin/resolwe', str(index)),
                [tool]
            ))

        volumes += self.tools_volumes

        # Create volumes for runtime (all read-only).
        runtime_volume_maps = SETTINGS.get('RUNTIME_VOLUME_MAPS', None)
        if runtime_volume_maps:
            for src, dst in runtime_volume_maps.items():
                volumes.append(new_volume(
                    'runtime',
                    'RUNTIME_DIR',
                    dst,
                    [self.data_id, src],
                ))

        # Add any extra volumes verbatim.
        volumes += SETTINGS.get('FLOW_DOCKER_EXTRA_VOLUMES', [])

        # Make sure that tmp dir exists.
        os.makedirs(constants.TMPDIR, mode=0o755, exist_ok=True)

        # Create Docker --volume parameters from volumes.
        command_args['volumes'] = ' '.join(['--volume="{src}":"{dest}":{options}'.format(**volume)
                                            for volume in volumes])

        # Set working directory to the data volume.
        command_args['workdir'] = '--workdir={}'.format(constants.DATA_VOLUME)

        # Change user inside the container.
        command_args['user'] = '--user={}:{}'.format(os.getuid(), os.getgid())

        # A non-login Bash shell should be used here (a subshell will be spawned later).
        command_args['shell'] = '/bin/bash'

        # Check if image exists locally. If not, command will exit with non-zero returncode
        check_command = '{command} image inspect {container_image}'.format(**command_args)

        logger.debug("Checking existence of docker image: {}".format(command_args['container_image']))

        check_proc = await subprocess.create_subprocess_exec(  # pylint: disable=no-member
            *shlex.split(check_command),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        await check_proc.communicate()

        if check_proc.returncode != 0:
            pull_command = '{command} pull {container_image}'.format(**command_args)

            logger.info("Pulling docker image: {}".format(command_args['container_image']))

            pull_proc = await subprocess.create_subprocess_exec(  # pylint: disable=no-member
                *shlex.split(pull_command),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE
            )
            _, stderr = await pull_proc.communicate()

            if pull_proc.returncode != 0:
                error_msg = "Docker failed to pull {} image.".format(command_args['container_image'])
                if stderr:
                    error_msg = '\n'.join([error_msg, stderr.decode('utf-8')])
                raise RuntimeError(error_msg)

        docker_command = (
            '{command} run --rm --interactive {container_name} {network} {volumes} {limits} '
            '{security} {workdir} {user} {container_image} {shell}'.format(**command_args)
        )

        logger.info("Starting docker container with command: {}".format(docker_command))
        start_time = time.time()

        # Workaround for pylint issue #1469
        # (https://github.com/PyCQA/pylint/issues/1469).
        self.proc = await subprocess.create_subprocess_exec(  # pylint: disable=no-member
            *shlex.split(docker_command),
            limit=4 * (2 ** 20),  # 4MB buffer size for line buffering
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT
        )

        stdout = []

        async def wait_for_container():
            """Wait for Docker container to start to avoid blocking the code that uses it."""
            self.proc.stdin.write(('echo PING' + os.linesep).encode('utf-8'))
            await self.proc.stdin.drain()
            while True:
                line = await self.proc.stdout.readline()
                stdout.append(line)
                if line.rstrip() == b'PING':
                    break
                if self.proc.stdout.at_eof():
                    raise RuntimeError()

        try:
            await asyncio.wait_for(wait_for_container(), timeout=DOCKER_START_TIMEOUT)
        except (asyncio.TimeoutError, RuntimeError):
            error_msg = "Docker container has not started for {} seconds.".format(DOCKER_START_TIMEOUT)
            stdout = ''.join([line.decode('utf-8') for line in stdout if line])
            if stdout:
                error_msg = '\n'.join([error_msg, stdout])
            raise RuntimeError(error_msg)

        end_time = time.time()
        logger.info("It took {:.2f}s for Docker container to start".format(end_time - start_time))

        self.stdout = self.proc.stdout

    async def run_script(self, script):
        """Execute the script and save results."""
        # Create a Bash command to add all the tools to PATH.
        tools_paths = ':'.join([map_["dest"] for map_ in self.tools_volumes])
        add_tools_path = 'export PATH=$PATH:{}'.format(tools_paths)
        # Spawn another child bash, to avoid running anything as PID 1, which has special
        # signal handling (e.g., cannot be SIGKILL-ed from inside).
        # A login Bash shell is needed to source /etc/profile.
        bash_line = '/bin/bash --login; exit $?' + os.linesep
        script = os.linesep.join(['set -x', 'set +B', add_tools_path, script]) + os.linesep
        self.proc.stdin.write(bash_line.encode('utf-8'))
        await self.proc.stdin.drain()
        self.proc.stdin.write(script.encode('utf-8'))
        await self.proc.stdin.drain()
        self.proc.stdin.close()

    async def end(self):
        """End process execution."""
        try:
            await self.proc.wait()
        finally:
            # Cleanup temporary files.
            for temporary_file in self.temporary_files:
                temporary_file.close()
            self.temporary_files = []

        return self.proc.returncode

    async def terminate(self):
        """Terminate a running script."""
        # Workaround for pylint issue #1469
        # (https://github.com/PyCQA/pylint/issues/1469).
        cmd = await subprocess.create_subprocess_exec(  # pylint: disable=no-member
            *shlex.split('{} rm -f {}'.format(self.command, self._generate_container_name()))
        )
        await cmd.wait()
        await self.proc.wait()

        await super().terminate()



""".. Ignore pydocstyle D400.

===========================
Elastic Management commands
===========================

Elastic app includes following Django management commands:

.. automodule:: resolwe.elastic.management.commands.elastic_index

.. automodule:: resolwe.elastic.management.commands.elastic_mapping

.. automodule:: resolwe.elastic.management.commands.elastic_purge

"""



"""Resolwe collection model."""
from django.contrib.postgres.fields import ArrayField, JSONField
from django.db import models

from .base import BaseModel
from .utils import DirtyError, validate_schema


class BaseCollection(BaseModel):
    """Template for Postgres model for storing a collection."""

    class Meta(BaseModel.Meta):
        """BaseCollection Meta options."""

        abstract = True

    #: detailed description
    description = models.TextField(blank=True)

    settings = JSONField(default=dict)

    data = models.ManyToManyField('flow.Data')

    #: collection descriptor schema
    descriptor_schema = models.ForeignKey('flow.DescriptorSchema', blank=True, null=True, on_delete=models.PROTECT)

    #: collection descriptor
    descriptor = JSONField(default=dict)

    #: indicate whether `descriptor` doesn't match `descriptor_schema` (is dirty)
    descriptor_dirty = models.BooleanField(default=False)

    #: tags for categorizing objects
    tags = ArrayField(models.CharField(max_length=255), default=list)

    def save(self, *args, **kwargs):
        """Perform descriptor validation and save object."""
        if self.descriptor_schema:
            try:
                validate_schema(self.descriptor, self.descriptor_schema.schema)  # pylint: disable=no-member
                self.descriptor_dirty = False
            except DirtyError:
                self.descriptor_dirty = True
        elif self.descriptor and self.descriptor != {}:
            raise ValueError("`descriptor_schema` must be defined if `descriptor` is given")

        super().save()


class Collection(BaseCollection):
    """Postgres model for storing a collection."""

    class Meta(BaseCollection.Meta):
        """Collection Meta options."""

        permissions = (
            ("view_collection", "Can view collection"),
            ("edit_collection", "Can edit collection"),
            ("share_collection", "Can share collection"),
            ("download_collection", "Can download files from collection"),
            ("add_collection", "Can add data objects to collection"),
            ("owner_collection", "Is owner of the collection"),
        )



""".. Ignore pydocstyle D400.

==========
Dispatcher
==========

"""
import asyncio
import inspect
import json
import logging
import os
import shlex
import shutil
from contextlib import suppress
from importlib import import_module

from asgiref.sync import async_to_sync
from channels.db import database_sync_to_async
from channels.exceptions import ChannelFull
from redis.exceptions import ConnectionError as RedisConnectionError

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.db import IntegrityError, connection, transaction
from django.db.models import Q

from resolwe.flow.engine import InvalidEngineError, load_engines
from resolwe.flow.execution_engines import ExecutionError
from resolwe.flow.models import Data, DataDependency, Process
from resolwe.test.utils import is_testing
from resolwe.utils import BraceMessage as __

from . import consumer, state
from .protocol import ExecutorFiles, WorkerProtocol

if settings.USE_TZ:
    from django.utils.timezone import now  # pylint: disable=ungrouped-imports
else:
    import datetime
    now = datetime.datetime.now  # pylint: disable=invalid-name

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

DEFAULT_CONNECTOR = 'resolwe.flow.managers.workload_connectors.local'


class SettingsJSONifier(json.JSONEncoder):
    """Customized JSON encoder, coercing all unknown types into strings.

    Needed due to the class hierarchy coming out of the database,
    which can't be serialized using the vanilla json encoder.
    """

    def default(self, o):  # pylint: disable=method-hidden
        """Try default; otherwise, coerce the object into a string."""
        try:
            return super().default(o)
        except TypeError:
            return str(o)


def dependency_status(data):
    """Return abstracted satus of dependencies.

    - ``STATUS_ERROR`` .. one dependency has error status or was deleted
    - ``STATUS_DONE`` .. all dependencies have done status
    - ``None`` .. other

    """
    parents_statuses = set(
        DataDependency.objects.filter(
            child=data, kind=DataDependency.KIND_IO
        ).distinct('parent__status').values_list('parent__status', flat=True)
    )

    if not parents_statuses:
        return Data.STATUS_DONE

    if None in parents_statuses:
        # Some parents have been deleted.
        return Data.STATUS_ERROR

    if Data.STATUS_ERROR in parents_statuses:
        return Data.STATUS_ERROR

    if len(parents_statuses) == 1 and Data.STATUS_DONE in parents_statuses:
        return Data.STATUS_DONE

    return None


class Manager:
    """The manager handles process job dispatching.

    Each :class:`~resolwe.flow.models.Data` object that's still waiting
    to be resolved is dispatched to a concrete workload management
    system (such as Celery or SLURM). The specific manager for that
    system (descended from
    :class:`~resolwe.flow.managers.workload_connectors.base.BaseConnector`)
    then handles actual job setup and submission. The job itself is an
    executor invocation; the executor then in turn sets up a safe and
    well-defined environment within the workload manager's task in which
    the process is finally run.
    """

    class _SynchronizationManager:
        """Context manager which enables transaction-like semantics.

        Once it is entered, exiting will not be possible until all
        executors have finished, including those that were already
        running when the context was entered.
        """

        def __init__(self):
            """Initialize state."""
            self.value = 0
            self.active = False
            self.condition = asyncio.Condition()
            self.tag_sequence = []

        async def __aenter__(self):
            """Begin synchronized execution context."""
            self.active = True
            return self

        async def __aexit__(self, exc_type, exc_val, exc_tb):
            """Wait for executors to finish, then return."""
            logger.info(__("Waiting for executor count to drop to 0, now it is {}", self.value))

            await self.condition.acquire()
            try:
                await self.condition.wait()
            finally:
                self.condition.release()
            logger.debug(__(
                "Sync semaphore dropped to 0, tag sequence was {}.",
                self.tag_sequence
            ))

            self.active = False
            return False

        async def reset(self):
            """Reset the semaphore to 0."""
            assert not self.active
            await self.condition.acquire()
            self.value = 0
            self.condition.release()

        async def inc(self, tag):
            """Increase executor count by 1."""
            await self.condition.acquire()
            self.value += 1
            logger.debug(__("Sync semaphore increased to {}, tag {}.", self.value, tag))
            self.tag_sequence.append(tag + '-up')
            self.condition.release()

        async def dec(self, tag):
            """Decrease executor count by 1.

            Return ``True`` if the count dropped to 0 as a result.
            """
            ret = False
            await self.condition.acquire()
            try:
                self.value -= 1
                logger.debug(__("Sync semaphore decreased to {}, tag {}.", self.value, tag))
                self.tag_sequence.append(tag + '-down')
                ret = self.value == 0
                if self.active and self.value == 0:
                    self.condition.notify_all()
            finally:
                self.condition.release()
            return ret

    class _SynchronizationManagerDummy:
        """Dummy synchronization manager implementation.

        This is a dummy placeholder variant of
        :class:`~resolwe.flow.managers.dispatcher.Manager._SynchronizationManager`,
        doing nothing. It's needed so that code async initialization can
        be done late enough, after code that already syntactically needs
        the synchronization manager, but doesn't need it to work yet.
        """

        def __init__(self):
            self.active = False
            self.value = 0

        async def inc(self, tag):  # pylint: disable=missing-docstring
            pass

        async def dec(self, tag):  # pylint: disable=missing-docstring
            pass

        async def reset(self):  # pylint: disable=missing-docstring
            pass

    class _SettingsManager:
        """Context manager for settings overrides.

        Because Django's :func:`~django.test.override_settings` is a
        context manager, it would make the code awkward if the manager's
        support for this wasn't also in the form of a context manager.
        """

        def __init__(self, state_key_prefix, **kwargs):
            """Prepare the context manager with the given overrides.

            :param state_key_prefix: The Redis key prefix used by
                :class:`~resolwe.flow.managers.state.ManagerState`.
            :param kwargs: The settings to override.
            """
            self.overrides = kwargs
            self.old_overrides = None
            self.state = state.ManagerState(state_key_prefix)

        def __enter__(self):
            """Begin context with overridden settings."""
            self.old_overrides = self.state.settings_override
            # Get a new copy of the settings, so we can modify them
            merged = self.state.settings_override or {}
            merged.update(self.overrides)
            self.state.settings_override = merged
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            """Clean up the override context."""
            self.state.settings_override = self.old_overrides
            return False

    def discover_engines(self, executor=None):
        """Discover configured engines.

        :param executor: Optional executor module override
        """
        if executor is None:
            executor = getattr(settings, 'FLOW_EXECUTOR', {}).get('NAME', 'resolwe.flow.executors.local')
        self.executor = self.load_executor(executor)
        logger.info(
            __("Loaded '{}' executor.", str(self.executor.__class__.__module__).replace('.prepare', ''))
        )

        expression_engines = getattr(settings, 'FLOW_EXPRESSION_ENGINES', ['resolwe.flow.expression_engines.jinja'])
        self.expression_engines = self.load_expression_engines(expression_engines)
        logger.info(__(
            "Found {} expression engines: {}", len(self.expression_engines), ', '.join(self.expression_engines.keys())
        ))

        execution_engines = getattr(settings, 'FLOW_EXECUTION_ENGINES', ['resolwe.flow.execution_engines.bash'])
        self.execution_engines = self.load_execution_engines(execution_engines)
        logger.info(__(
            "Found {} execution engines: {}", len(self.execution_engines), ', '.join(self.execution_engines.keys())
        ))

    def reset(self, keep_state=False):
        """Reset the shared state and drain Django Channels.

        :param keep_state: If ``True``, do not reset the shared manager
            state (useful in tests, where the settings overrides need to
            be kept). Defaults to ``False``.
        """
        if not keep_state:
            self.state = state.ManagerState(state.MANAGER_STATE_PREFIX)
            self.state.reset()
        async_to_sync(consumer.run_consumer)(timeout=1)
        async_to_sync(self.sync_counter.reset)()

    def __init__(self, *args, **kwargs):
        """Initialize arguments."""
        self.discover_engines()
        self.state = state.ManagerState(state.MANAGER_STATE_PREFIX)

        # Don't call the full self.reset() here, that's only meant for testing
        # since it also runs a dummy consumer to drain channels.
        with suppress(RedisConnectionError):
            # It's awkward to handle documentation and migration testing
            # any other way.
            self.state.reset()

        # The number of executors currently running; used for test synchronization.
        # We need to start out with a dummy object, so that the async
        # infrastructure isn't started too early. In particular, this handles
        # counter functions that are called before any actual synchronization
        # is wanted: in a test, what's called first is the Django signal.
        # This will call communicate(), which will (has to) first try upping
        # the counter value; the future that that produces can be scheduled
        # to the wrong event loop (the application one is started further
        # down communicate() in the synchronization block), which then
        # leads to exceedingly obscure crashes further down the line.
        self.sync_counter = self._SynchronizationManagerDummy()

        # Django's override_settings should be avoided at all cost here
        # to keep the manager as independent as possible, and in
        # particular to avoid overriding with dangerous variables, such
        # as the one controlling Django signal synchronicity. To make
        # such avoidance possible, all settings lookups in the worker
        # should use a private dictionary instead of the global
        # configuration. This variable is maintained around
        # _data_scan() calls, effectively emulating Django's
        # settings but having no effect on anything outside the worker
        # code (in particular, the signal triggers).
        self.settings_actual = {}

        # Ensure there is only one manager instance per process. This
        # is required as other parts of the code access the global
        # manager instance.
        with suppress(ImportError):
            from resolwe.flow import managers
            assert not hasattr(managers, 'manager')

        self.scheduling_class_map = dict(Process.SCHEDULING_CLASS_CHOICES)

        # Check settings for consistency.
        flow_manager = getattr(settings, 'FLOW_MANAGER', {})
        if 'DISPATCHER_MAPPING' in flow_manager and 'NAME' in flow_manager:
            raise ImproperlyConfigured("Key 'DISPATCHER_MAPPING' conflicts with key 'NAME' in FLOW_MANAGER settings.")

        if 'DISPATCHER_MAPPING' in flow_manager:
            mapping = flow_manager['DISPATCHER_MAPPING']

            scheduling_classes = set(self.scheduling_class_map.values())
            map_keys = set(mapping.keys())
            class_difference = scheduling_classes - map_keys
            if class_difference:
                raise ImproperlyConfigured(
                    "Dispatcher manager mapping settings incomplete, missing {}.".format(class_difference)
                )
            connector_list = [mapping[klass] for klass in scheduling_classes]
        else:
            connector_list = [flow_manager.get('NAME', DEFAULT_CONNECTOR)]

        # Pre-load all needed connectors.
        self.connectors = {}
        for module_name in connector_list:
            connector_module = import_module(module_name)
            self.connectors[module_name] = connector_module.Connector()

        logger.info(__(
            "Found {} workload connectors: {}", len(self.connectors), ', '.join(self.connectors.keys())
        ))

        super().__init__(*args, **kwargs)

    def _marshal_settings(self):
        """Marshal Django settings into a serializable object.

        :return: The serialized settings.
        :rtype: dict
        """
        result = {}
        for key in dir(settings):
            if any(map(key.startswith, ['FLOW_', 'RESOLWE_', 'CELERY_'])):
                result[key] = getattr(settings, key)
        return result

    def _include_environment_variables(self, program, executor_vars):
        """Define environment variables."""
        env_vars = {
            'RESOLWE_HOST_URL': self.settings_actual.get('RESOLWE_HOST_URL', 'localhost'),
        }

        set_env = self.settings_actual.get('FLOW_EXECUTOR', {}).get('SET_ENV', {})
        env_vars.update(executor_vars)
        env_vars.update(set_env)

        export_commands = ['export {}={}'.format(key, shlex.quote(value)) for key, value in env_vars.items()]
        return os.linesep.join(export_commands) + os.linesep + program

    def run(self, data, runtime_dir, argv):
        """Select a concrete connector and run the process through it.

        :param data: The :class:`~resolwe.flow.models.Data` object that
            is to be run.
        :param runtime_dir: The directory the executor is run from.
        :param argv: The argument vector used to spawn the executor.
        """
        process_scheduling = self.scheduling_class_map[data.process.scheduling_class]
        if 'DISPATCHER_MAPPING' in getattr(settings, 'FLOW_MANAGER', {}):
            class_name = settings.FLOW_MANAGER['DISPATCHER_MAPPING'][process_scheduling]
        else:
            class_name = getattr(settings, 'FLOW_MANAGER', {}).get('NAME', DEFAULT_CONNECTOR)

        data.scheduled = now()
        data.save(update_fields=['scheduled'])

        async_to_sync(self.sync_counter.inc)('executor')
        return self.connectors[class_name].submit(data, runtime_dir, argv)

    def _get_per_data_dir(self, dir_base, data_id):
        """Extend the given base directory with a per-data component.

        The method creates a private path for the
        :class:`~resolwe.flow.models.Data` object, such as::

            ./test_data/1/

        if ``base_dir`` is ``'./test_data'`` and ``data_id`` is ``1``.

        :param dir_base: The base path to be extended. This will usually
            be one of the directories configured in the
            ``FLOW_EXECUTOR`` setting.
        :param data_id: The :class:`~resolwe.flow.models.Data` object's
            id used for the extending.
        :return: The new path for the :class:`~resolwe.flow.models.Data`
            object.
        :rtype: str
        """
        # Use Django settings here, because the state must be preserved
        # across events. This also implies the directory settings can't
        # be patched outside the manager and then just sent along in the
        # command packets.
        result = self.settings_actual.get('FLOW_EXECUTOR', {}).get(dir_base, '')
        return os.path.join(result, str(data_id))

    def _prepare_data_dir(self, data_id):
        """Prepare destination directory where the data will live.

        :param data_id: The :class:`~resolwe.flow.models.Data` object id
            for which to prepare the private execution directory.
        :return: The prepared data directory path.
        :rtype: str
        """
        logger.debug(__("Preparing data directory for Data with id {}.", data_id))

        output_path = self._get_per_data_dir('DATA_DIR', data_id)
        dir_mode = self.settings_actual.get('FLOW_EXECUTOR', {}).get('DATA_DIR_MODE', 0o755)
        os.mkdir(output_path, mode=dir_mode)
        # os.mkdir is not guaranteed to set the given mode
        os.chmod(output_path, dir_mode)
        return output_path

    def _prepare_context(self, data_id, data_dir, runtime_dir, **kwargs):
        """Prepare settings and constants JSONs for the executor.

        Settings and constants provided by other ``resolwe`` modules and
        :class:`~django.conf.settings` are all inaccessible in the
        executor once it is deployed, so they need to be serialized into
        the runtime directory.

        :param data_id: The :class:`~resolwe.flow.models.Data` object id
            being prepared for.
        :param data_dir: The target execution directory for this
            :class:`~resolwe.flow.models.Data` object.
        :param runtime_dir: The target runtime support directory for
            this :class:`~resolwe.flow.models.Data` object; this is
            where the environment is serialized into.
        :param kwargs: Extra settings to include in the main settings
            file.
        """
        files = {}
        secrets = {}

        settings_dict = {}
        settings_dict['DATA_DIR'] = data_dir
        settings_dict['REDIS_CHANNEL_PAIR'] = state.MANAGER_EXECUTOR_CHANNELS
        files[ExecutorFiles.EXECUTOR_SETTINGS] = settings_dict

        django_settings = {}
        django_settings.update(self.settings_actual)
        django_settings.update(kwargs)
        files[ExecutorFiles.DJANGO_SETTINGS] = django_settings

        # Add scheduling classes.
        files[ExecutorFiles.PROCESS_META] = {
            k: getattr(Process, k) for k in dir(Process)
            if k.startswith('SCHEDULING_CLASS_') and isinstance(getattr(Process, k), str)
        }

        # Add Data status constants.
        files[ExecutorFiles.DATA_META] = {
            k: getattr(Data, k) for k in dir(Data)
            if k.startswith('STATUS_') and isinstance(getattr(Data, k), str)
        }

        # Extend the settings with whatever the executor wants.
        self.executor.extend_settings(data_id, files, secrets)

        # Save the settings into the various files in the runtime dir.
        settings_dict[ExecutorFiles.FILE_LIST_KEY] = list(files.keys())
        for file_name in files:
            file_path = os.path.join(runtime_dir, file_name)
            with open(file_path, 'wt') as json_file:
                json.dump(files[file_name], json_file, cls=SettingsJSONifier)

        # Save the secrets in the runtime dir, with permissions to prevent listing the given
        # directory.
        secrets_dir = os.path.join(runtime_dir, ExecutorFiles.SECRETS_DIR)
        os.makedirs(secrets_dir, mode=0o300)
        for file_name, value in secrets.items():
            file_path = os.path.join(secrets_dir, file_name)

            # Set umask to 0 to ensure that we set the correct permissions.
            old_umask = os.umask(0)
            try:
                # We need to use os.open in order to correctly enforce file creation. Otherwise,
                # there is a race condition which can be used to create the file with different
                # ownership/permissions.
                file_descriptor = os.open(file_path, os.O_WRONLY | os.O_CREAT, mode=0o600)
                with os.fdopen(file_descriptor, 'w') as raw_file:
                    raw_file.write(value)
            finally:
                os.umask(old_umask)

    def _prepare_executor(self, data_id, executor):
        """Copy executor sources into the destination directory.

        :param data_id: The :class:`~resolwe.flow.models.Data` object id
            being prepared for.
        :param executor: The fully qualified name of the executor that
            is to be used for this data object.
        :return: Tuple containing the relative fully qualified name of
            the executor class ('relative' to how the executor will be
            run) and the path to the directory where the executor will
            be deployed.
        :rtype: (str, str)
        """
        logger.debug(__("Preparing executor for Data with id {}", data_id))

        # Both of these imports are here only to get the packages' paths.
        import resolwe.flow.executors as executor_package

        exec_dir = os.path.dirname(inspect.getsourcefile(executor_package))
        dest_dir = self._get_per_data_dir('RUNTIME_DIR', data_id)
        dest_package_dir = os.path.join(dest_dir, 'executors')
        shutil.copytree(exec_dir, dest_package_dir)
        dir_mode = self.settings_actual.get('FLOW_EXECUTOR', {}).get('RUNTIME_DIR_MODE', 0o755)
        os.chmod(dest_dir, dir_mode)

        class_name = executor.rpartition('.executors.')[-1]
        return '.{}'.format(class_name), dest_dir

    def _prepare_script(self, dest_dir, program):
        """Copy the script into the destination directory.

        :param dest_dir: The target directory where the script will be
            saved.
        :param program: The script text to be saved.
        :return: The name of the script file.
        :rtype: str
        """
        script_name = ExecutorFiles.PROCESS_SCRIPT
        dest_file = os.path.join(dest_dir, script_name)
        with open(dest_file, 'wt') as dest_file_obj:
            dest_file_obj.write(program)
        os.chmod(dest_file, 0o700)
        return script_name

    def override_settings(self, **kwargs):
        """Override global settings within the calling context.

        :param kwargs: The settings overrides. Same use as for
            :func:`django.test.override_settings`.
        """
        return self._SettingsManager(self.state.key_prefix, **kwargs)

    async def handle_control_event(self, message):
        """Handle an event from the Channels layer.

        Channels layer callback, do not call directly.
        """
        cmd = message[WorkerProtocol.COMMAND]
        logger.debug(__("Manager worker got channel command '{}'.", cmd))

        # Prepare settings for use; Django overlaid by state overlaid by
        # anything immediate in the current packet.
        immediates = {}
        if cmd == WorkerProtocol.COMMUNICATE:
            immediates = message.get(WorkerProtocol.COMMUNICATE_SETTINGS, {}) or {}
        override = self.state.settings_override or {}
        override.update(immediates)
        self.settings_actual = self._marshal_settings()
        self.settings_actual.update(override)

        if cmd == WorkerProtocol.COMMUNICATE:
            try:
                await database_sync_to_async(self._data_scan)(**message[WorkerProtocol.COMMUNICATE_EXTRA])
            except Exception:  # pylint: disable=broad-except
                logger.exception("Unknown error occured while processing communicate control command.")
                raise
            finally:
                await self.sync_counter.dec('communicate')

        elif cmd == WorkerProtocol.FINISH:
            try:
                data_id = message[WorkerProtocol.DATA_ID]
                if not getattr(settings, 'FLOW_MANAGER_KEEP_DATA', False):
                    try:
                        def handle_error(func, path, exc_info):
                            """Handle permission errors while removing data directories."""
                            if isinstance(exc_info[1], PermissionError):
                                os.chmod(path, 0o700)
                                shutil.rmtree(path)

                        # Remove secrets directory, but leave the rest of the runtime directory
                        # intact. Runtime directory will be removed during data purge, when the
                        # data object is removed.
                        secrets_dir = os.path.join(
                            self._get_per_data_dir('RUNTIME_DIR', data_id),
                            ExecutorFiles.SECRETS_DIR
                        )
                        shutil.rmtree(secrets_dir, onerror=handle_error)
                    except OSError:
                        logger.exception("Manager exception while removing data runtime directory.")

                if message[WorkerProtocol.FINISH_SPAWNED]:
                    await database_sync_to_async(self._data_scan)(**message[WorkerProtocol.FINISH_COMMUNICATE_EXTRA])
            except Exception:  # pylint: disable=broad-except
                logger.exception(
                    "Unknown error occured while processing finish control command.",
                    extra={'data_id': data_id}
                )
                raise
            finally:
                await self.sync_counter.dec('executor')

        elif cmd == WorkerProtocol.ABORT:
            await self.sync_counter.dec('executor')

        else:
            logger.error(__("Ignoring unknown manager control command '{}'.", cmd))

    def _ensure_counter(self):
        """Ensure the sync counter is a valid non-dummy object."""
        if not isinstance(self.sync_counter, self._SynchronizationManager):
            self.sync_counter = self._SynchronizationManager()

    async def execution_barrier(self):
        """Wait for executors to finish.

        At least one must finish after this point to avoid a deadlock.
        """
        async def _barrier():
            """Enter the sync block and exit the app afterwards."""
            async with self.sync_counter:
                pass
            await consumer.exit_consumer()

        self._ensure_counter()
        await asyncio.wait([
            _barrier(),
            consumer.run_consumer(),
        ])
        self.sync_counter = self._SynchronizationManagerDummy()

    async def communicate(self, data_id=None, run_sync=False, save_settings=True):
        """Scan database for resolving Data objects and process them.

        This is submitted as a task to the manager's channel workers.

        :param data_id: Optional id of Data object which (+ its
            children) should be processes. If it is not given, all
            resolving objects are processed.
        :param run_sync: If ``True``, wait until all processes spawned
            from this point on have finished processing. If no processes
            are spawned, this results in a deadlock, since counts are
            handled on process finish.
        :param save_settings: If ``True``, save the current Django
            settings context to the global state. This should never be
            ``True`` for "automatic" calls, such as from Django signals,
            which can be invoked from inappropriate contexts (such as in
            the listener). For user code, it should be left at the
            default value. The saved settings are in effect until the
            next such call.
        """
        executor = getattr(settings, 'FLOW_EXECUTOR', {}).get('NAME', 'resolwe.flow.executors.local')
        logger.debug(__(
            "Manager sending communicate command on '{}' triggered by Data with id {}.",
            state.MANAGER_CONTROL_CHANNEL,
            data_id,
        ))

        saved_settings = self.state.settings_override
        if save_settings:
            saved_settings = self._marshal_settings()
            self.state.settings_override = saved_settings

        if run_sync:
            self._ensure_counter()
        await self.sync_counter.inc('communicate')
        try:
            await consumer.send_event({
                WorkerProtocol.COMMAND: WorkerProtocol.COMMUNICATE,
                WorkerProtocol.COMMUNICATE_SETTINGS: saved_settings,
                WorkerProtocol.COMMUNICATE_EXTRA: {
                    'data_id': data_id,
                    'executor': executor,
                },
            })
        except ChannelFull:
            logger.exception("ChannelFull error occurred while sending communicate message.")
            await self.sync_counter.dec('communicate')

        if run_sync and not self.sync_counter.active:
            logger.debug(__(
                "Manager on channel '{}' entering synchronization block.",
                state.MANAGER_CONTROL_CHANNEL
            ))
            await self.execution_barrier()
            logger.debug(__(
                "Manager on channel '{}' exiting synchronization block.",
                state.MANAGER_CONTROL_CHANNEL
            ))

    def _data_execute(self, data, program, executor):
        """Execute the Data object.

        The activities carried out here include target directory
        preparation, executor copying, setting serialization and actual
        execution of the object.

        :param data: The :class:`~resolwe.flow.models.Data` object to
            execute.
        :param program: The process text the manager got out of
            execution engine evaluation.
        :param executor: The executor to use for this object.
        """
        if not program:
            return

        logger.debug(__("Manager preparing Data with id {} for processing.", data.id))

        # Prepare the executor's environment.
        try:
            executor_env_vars = self.get_executor().get_environment_variables()
            program = self._include_environment_variables(program, executor_env_vars)
            data_dir = self._prepare_data_dir(data.id)
            executor_module, runtime_dir = self._prepare_executor(data.id, executor)

            # Execute execution engine specific runtime preparation.
            execution_engine = data.process.run.get('language', None)
            volume_maps = self.get_execution_engine(execution_engine).prepare_runtime(runtime_dir, data)

            self._prepare_context(data.id, data_dir, runtime_dir, RUNTIME_VOLUME_MAPS=volume_maps)
            self._prepare_script(runtime_dir, program)

            argv = [
                '/bin/bash',
                '-c',
                self.settings_actual.get('FLOW_EXECUTOR', {}).get('PYTHON', '/usr/bin/env python')
                + ' -m executors ' + executor_module
            ]
        except PermissionDenied as error:
            data.status = Data.STATUS_ERROR
            data.process_error.append("Permission denied for process: {}".format(error))
            data.save()
            return
        except OSError as err:
            logger.error(__(
                "OSError occurred while preparing data {} (will skip): {}",
                data.id, err
            ))
            return

        # Hand off to the run() method for execution.
        logger.info(__("Running {}", runtime_dir))
        self.run(data, runtime_dir, argv)

    def _data_scan(self, data_id=None, executor='resolwe.flow.executors.local', **kwargs):
        """Scan for new Data objects and execute them.

        :param data_id: Optional id of Data object which (+ its
            children) should be scanned. If it is not given, all
            resolving objects are processed.
        :param executor: The fully qualified name of the executor to use
            for all :class:`~resolwe.flow.models.Data` objects
            discovered in this pass.
        """
        def process_data_object(data):
            """Process a single data object."""
            # Lock for update. Note that we want this transaction to be as short as possible in
            # order to reduce contention and avoid deadlocks. This is why we do not lock all
            # resolving objects for update, but instead only lock one object at a time. This
            # allows managers running in parallel to process different objects.
            data = Data.objects.select_for_update().get(pk=data.pk)
            if data.status != Data.STATUS_RESOLVING:
                # The object might have already been processed while waiting for the lock to be
                # obtained. In this case, skip the object.
                return

            dep_status = dependency_status(data)

            if dep_status == Data.STATUS_ERROR:
                data.status = Data.STATUS_ERROR
                data.process_error.append("One or more inputs have status ERROR")
                data.process_rc = 1
                data.save()
                return

            elif dep_status != Data.STATUS_DONE:
                return

            if data.process.run:
                try:
                    execution_engine = data.process.run.get('language', None)
                    # Evaluation by the execution engine may spawn additional data objects and
                    # perform other queries on the database. Queries of all possible execution
                    # engines need to be audited for possibilities of deadlocks in case any
                    # additional locks are introduced. Currently, we only take an explicit lock on
                    # the currently processing object.
                    program = self.get_execution_engine(execution_engine).evaluate(data)
                except (ExecutionError, InvalidEngineError) as error:
                    data.status = Data.STATUS_ERROR
                    data.process_error.append("Error in process script: {}".format(error))
                    data.save()
                    return

                # Set allocated resources:
                resource_limits = data.process.get_resource_limits()
                data.process_memory = resource_limits['memory']
                data.process_cores = resource_limits['cores']
            else:
                # If there is no run section, then we should not try to run anything. But the
                # program must not be set to None as then the process will be stuck in waiting
                # state.
                program = ''

            if data.status != Data.STATUS_DONE:
                # The data object may already be marked as done by the execution engine. In this
                # case we must not revert the status to STATUS_WAITING.
                data.status = Data.STATUS_WAITING
            data.save(render_name=True)

            # Actually run the object only if there was nothing with the transaction.
            transaction.on_commit(
                # Make sure the closure gets the right values here, since they're
                # changed in the loop.
                lambda d=data, p=program: self._data_execute(d, p, executor)
            )

        logger.debug(__("Manager processing communicate command triggered by Data with id {}.", data_id))

        if is_testing():
            # NOTE: This is a work-around for Django issue #10827
            # (https://code.djangoproject.com/ticket/10827), same as in
            # TestCaseHelpers._pre_setup(). Because the worker is running
            # independently, it must clear the cache on its own.
            ContentType.objects.clear_cache()

            # Ensure settings overrides apply
            self.discover_engines(executor=executor)

        try:
            queryset = Data.objects.filter(status=Data.STATUS_RESOLVING)
            if data_id is not None:
                # Scan only given data object and its children.
                queryset = queryset.filter(Q(parents=data_id) | Q(id=data_id)).distinct()

            for data in queryset:
                try:
                    with transaction.atomic():
                        process_data_object(data)

                        # All data objects created by the execution engine are commited after this
                        # point and may be processed by other managers running in parallel. At the
                        # same time, the lock for the current data object is released.
                except Exception as error:  # pylint: disable=broad-except
                    logger.exception(__(
                        "Unhandled exception in _data_scan while processing data object {}.",
                        data.pk
                    ))

                    # Unhandled error while processing a data object. We must set its
                    # status to STATUS_ERROR to prevent the object from being retried
                    # on next _data_scan run. We must perform this operation without
                    # using the Django ORM as using the ORM may be the reason the error
                    # occurred in the first place.
                    error_msg = "Internal error: {}".format(error)
                    process_error_field = Data._meta.get_field('process_error')  # pylint: disable=protected-access
                    max_length = process_error_field.base_field.max_length
                    if len(error_msg) > max_length:
                        error_msg = error_msg[:max_length - 3] + '...'

                    try:
                        with connection.cursor() as cursor:
                            cursor.execute(
                                """
                                    UPDATE {table}
                                    SET
                                        status = %(status)s,
                                        process_error = process_error || (%(error)s)::varchar[]
                                    WHERE id = %(id)s
                                """.format(
                                    table=Data._meta.db_table  # pylint: disable=protected-access
                                ),
                                {
                                    'status': Data.STATUS_ERROR,
                                    'error': [error_msg],
                                    'id': data.pk
                                }
                            )
                    except Exception as error:  # pylint: disable=broad-except
                        # If object's state cannot be changed due to some database-related
                        # issue, at least skip the object for this run.
                        logger.exception(__(
                            "Unhandled exception in _data_scan while trying to emit error for {}.",
                            data.pk
                        ))

        except IntegrityError as exp:
            logger.error(__("IntegrityError in manager {}", exp))
            return

    def get_executor(self):
        """Return an executor instance."""
        return self.executor

    def get_expression_engine(self, name):
        """Return an expression engine instance."""
        try:
            return self.expression_engines[name]
        except KeyError:
            raise InvalidEngineError("Unsupported expression engine: {}".format(name))

    def get_execution_engine(self, name):
        """Return an execution engine instance."""
        try:
            return self.execution_engines[name]
        except KeyError:
            raise InvalidEngineError("Unsupported execution engine: {}".format(name))

    def load_executor(self, executor_name):
        """Load process executor."""
        executor_name = executor_name + '.prepare'
        module = import_module(executor_name)
        return module.FlowExecutorPreparer()

    def load_expression_engines(self, engines):
        """Load expression engines."""
        return load_engines(self, 'ExpressionEngine', 'expression_engines', engines)

    def load_execution_engines(self, engines):
        """Load execution engines."""
        return load_engines(self, 'ExecutionEngine', 'execution_engines', engines)



""".. Ignore pydocstyle D400.

=======
Elastic
=======

Framework for advanced indexing of Django models with ElasticSearch.

To register index processor, create `elastic_indexes.py` file int your
app and put subclass of :class:`~resolwe.elastic.indices.BaseIndex` in
it. It will automatically register and index all objects specified in
it.

For building the index for the first time or manually updating it, run::

    python manage.py elastic_index

.. automodule:: resolwe.elastic.indices
.. automodule:: resolwe.elastic.viewsets
.. automodule:: resolwe.elastic.builder
.. automodule:: resolwe.elastic.pagination
.. automodule:: resolwe.elastic.utils
.. automodule:: resolwe.elastic.management.commands

"""
from elasticsearch_dsl.connections import connections

default_app_config = 'resolwe.elastic.apps.ElasticConfig'  # pylint: disable=invalid-name



"""Resolwe relation serializer."""
from django.db import transaction

from rest_framework import serializers

from resolwe.flow.models.collection import Collection
from resolwe.flow.models.entity import Relation, RelationPartition, RelationType
from resolwe.rest.serializers import SelectiveFieldMixin

from .base import ResolweBaseSerializer


class RelationPartitionSerializer(SelectiveFieldMixin, serializers.ModelSerializer):
    """Serializer for RelationPartition objects."""

    class Meta:
        """RelationPartitionSerializer Meta options."""

        model = RelationPartition
        fields = ('id', 'entity', 'position', 'label')


class RelationSerializer(ResolweBaseSerializer):
    """Serializer for Relation objects."""

    partitions = RelationPartitionSerializer(source='relationpartition_set', many=True)
    collection = serializers.PrimaryKeyRelatedField(queryset=Collection.objects.all())
    type = serializers.SlugRelatedField(queryset=RelationType.objects.all(), slug_field='name')

    class Meta:
        """RelationSerializer Meta options."""

        model = Relation
        read_only_fields = (
            'created',
            'id',
            'modified',
        )
        update_protected_fields = (
            'contributor',
            'type',
        )
        fields = read_only_fields + update_protected_fields + (
            'collection',
            'category',
            'partitions',
            'unit',
        )

    def validate_partitions(self, partitions):
        """Raise validation error if list of partitions is empty."""
        if not partitions:
            raise serializers.ValidationError("List of partitions must not be empty.")

        return partitions

    def _create_partitions(self, instance, partitions):
        """Create partitions."""
        for partition in partitions:
            RelationPartition.objects.create(
                relation=instance,
                entity=partition['entity'],
                label=partition.get('label', None),
                position=partition.get('position', None),
            )

    def create(self, validated_data):
        """Create ``Relation`` object and add partitions of ``Entities``."""
        # `partitions` field is renamed to `relationpartition_set` based on source of nested serializer
        partitions = validated_data.pop('relationpartition_set')

        with transaction.atomic():
            instance = Relation.objects.create(**validated_data)
            self._create_partitions(instance, partitions)

        return instance

    def update(self, instance, validated_data):
        """Update ``Relation``."""
        # `partitions` field is renamed to `relationpartition_set` based on source of nested serializer
        partitions = validated_data.pop('relationpartition_set', None)

        with transaction.atomic():
            instance = super().update(instance, validated_data)

            if partitions is not None:
                # TODO: Apply the diff instead of recreating all objects.
                instance.relationpartition_set.all().delete()
                self._create_partitions(instance, partitions)

        return instance



# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2017-10-19 09:37
from __future__ import unicode_literals

from django.db import migrations

from resolwe.flow.models.utils import hydrate_size
from resolwe.flow.utils import iterate_fields


def calculate_total_size(apps, schema_editor):
    """Add ``total_size`` field to all file/dir-type outputs."""
    Data = apps.get_model('flow', 'Data')
    for data in Data.objects.all():
        hydrate_size(data, force=True)
        data.save()


def remove_total_size(apps, schema_editor):
    """Remove ``total_size`` field from all file/dir-type outputs."""
    Data = apps.get_model('flow', 'Data')
    for data in Data.objects.all():
        for field_schema, fields in iterate_fields(data.output, data.process.output_schema):
            name = field_schema['name']
            value = fields[name]
            if 'type' in field_schema:
                if field_schema['type'].startswith('basic:file:'):
                    del value['total_size']
                elif field_schema['type'].startswith('list:basic:file:'):
                    for obj in value:
                        del obj['total_size']
                elif field_schema['type'].startswith('basic:dir:'):
                    del value['total_size']
                elif field_schema['type'].startswith('list:basic:dir:'):
                    for obj in value:
                        del obj['total_size']
        data.save()


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0005_data_dependency_3'),
    ]

    operations = [
        migrations.RunPython(calculate_total_size, remove_total_size),
    ]



""".. Ignore pydocstyle D400.

=========================
Delete Unreferenced Files
=========================

"""
from django.core.management.base import BaseCommand

from resolwe.flow.utils.purge import data_purge


class Command(BaseCommand):
    """Purge files with no reference in Data objects."""

    help = "Purge files with no reference in Data objects."

    def add_arguments(self, parser):
        """Command arguments."""
        parser.add_argument('-d', '--data', type=str, nargs='+', help="list of data ids")
        parser.add_argument('-f', '--force', action='store_true', help="delete unreferenced files")

    def handle(self, *args, **options):
        """Call :func:`~resolwe.flow.utils.purge.data_purge`."""
        data_purge(options['data'], options['force'], options['verbosity'])



"""Process runtime."""
import logging

import resolwe_runtime_utils

from .descriptor import *  # pylint: disable=wildcard-import,unused-wildcard-import
from .fields import *  # pylint: disable=wildcard-import,unused-wildcard-import

try:
    from plumbum import local as Cmd  # pylint: disable=unused-import
except ImportError:
    # Generate a dummy class that defers the ImportError to a point where
    # the package actually tries to be used.
    class MissingPlumbumCmd:
        """Missing plumbum command dummy class."""

        def __getitem__(self, cmd):
            """Raise ImportError exception."""
            raise ImportError("Missing 'plumbum' Python package.")

    Cmd = MissingPlumbumCmd()

# Inputs class name.
PROCESS_INPUTS_NAME = 'Input'
# Outputs class name.
PROCESS_OUTPUTS_NAME = 'Output'


class _IoBase:
    """Base class for input/output values."""

    frozen = False

    def __init__(self, **kwargs):
        """Construct an I/O object."""
        super().__setattr__('_data', {})
        self._data.update(kwargs)

    def freeze(self):
        """Freeze the object, preventing further mutation."""
        super().__setattr__('frozen', True)
        return self

    def items(self):
        """Return iterator over I/O items."""
        return self._data.items()

    def __getattr__(self, name):
        try:
            return self._data[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        if self.frozen:
            raise ValueError("cannot mutate frozen object")

        self._data[name] = value

    def __delattr__(self, name):
        if self.frozen:
            raise ValueError("cannot mutate frozen object")

        try:
            del self._data[name]
        except KeyError:
            raise AttributeError(name)

    def __repr__(self):
        return '<{} {}>'.format(
            self.__class__.__name__,
            ', '.join([
                '{}={}'.format(field, repr(value))
                for field, value in self._data.items()
            ]),
        )


class Inputs(_IoBase):
    """Process inputs."""

    pass


class Outputs(_IoBase):
    """Process outputs."""

    pass


class ProcessMeta(type):
    """Metaclass for processes during runtime."""

    def __new__(mcs, name, bases, namespace, **kwargs):
        """Create new process class."""
        if namespace['__module__'] == 'resolwe.process.runtime':
            return type.__new__(mcs, name, bases, namespace)

        # Generate a runtime version of the process descriptor.
        meta = ProcessDescriptor()
        for meta_name in dir(meta.metadata):
            if meta_name.startswith('_'):
                continue

            setattr(meta.metadata, meta_name, namespace.pop(meta_name, None))

        for nsp, fields in ((PROCESS_INPUTS_NAME, meta.inputs), (PROCESS_OUTPUTS_NAME, meta.outputs)):
            inputs = namespace.pop(nsp, {})
            for field_name in dir(inputs):
                if field_name.startswith('_'):
                    continue

                field = getattr(inputs, field_name)
                if not isinstance(field, Field):
                    continue
                field.contribute_to_class(meta, fields, field_name)

        meta.validate()

        result = type.__new__(mcs, name, bases, namespace)
        result._meta = meta  # pylint: disable=protected-access
        return result


class Process(metaclass=ProcessMeta):
    """Resolwe process."""

    def __init__(self):
        """Construct a new process instance."""
        self.logger = logging.getLogger(self.__class__.__name__)

    def run(self, inputs, outputs):
        """Process entry point."""
        raise NotImplementedError

    def progress(self, progress):
        """Report process progress.

        :param progress: A float between 0 and 1 denoting the progress
        """
        report = resolwe_runtime_utils.progress(progress)
        # TODO: Use the protocol to report progress.
        print(report)

    def info(self, *args):
        """Log informational message."""
        report = resolwe_runtime_utils.info(' '.join([str(x) for x in args]))
        # TODO: Use the protocol to report progress.
        print(report)

    def warning(self, *args):
        """Log warning message."""
        report = resolwe_runtime_utils.warning(' '.join([str(x) for x in args]))
        # TODO: Use the protocol to report progress.
        print(report)

    def error(self, *args):
        """Log error message."""
        report = resolwe_runtime_utils.error(' '.join([str(x) for x in args]))
        # TODO: Use the protocol to report progress.
        print(report)

    def update_descriptor(self, **kwargs):
        """Update sample descriptor."""
        raise NotImplementedError

    def start(self, inputs):
        """Start the process.

        :param inputs: An instance of `Inputs` describing the process inputs
        :return: An instance of `Outputs` describing the process outputs
        """
        # pylint: disable=logging-format-interpolation
        self.logger.info("Process is starting")

        outputs = Outputs()

        # Validate and clean inputs.
        try:
            inputs = self._validate_inputs(inputs)
        except ValidationError as error:
            self.logger.error("Input validation has failed: {}".format(error.args[0]))
            raise

        self.logger.info("Process is running")
        try:
            self.run(inputs.freeze(), outputs)
            try:
                outputs = self._validate_outputs(outputs)
            except ValidationError as error:
                self.logger.error("Output validation has failed: {}".format(error.args[0]))
                raise

            return outputs.freeze()
        except:  # noqa pylint: disable=bare-except
            self.logger.exception("Exception while running process")
            raise
        finally:
            self.logger.info("Process has finished")

    def _validate_inputs(self, inputs):
        """Validate process inputs."""
        valid_inputs = Inputs()
        for name, field in self._meta.inputs.items():
            value = getattr(inputs, name, None)
            setattr(valid_inputs, name, field.clean(value))

        return valid_inputs

    def _validate_outputs(self, outputs):
        """Validate process outputs."""
        valid_outputs = Outputs()
        for name, field in self._meta.outputs.items():
            value = getattr(outputs, name, None)
            value = field.clean(value)
            value = field.to_output(value)
            setattr(valid_outputs, name, value)

        return valid_outputs



"""Search router."""
from rest_framework.routers import DefaultRouter, Route


class SearchRouter(DefaultRouter):
    """Custom router for search endpoints.

    Search endpoints don't follow REST principles and thus don't need
    routes that default router provides.

    """

    routes = [
        Route(
            url=r'^{prefix}{trailing_slash}$',
            mapping={
                'get': 'list',
                'post': 'list_with_post'
            },
            name='{basename}',
            initkwargs={}
        )
    ]



""".. Ignore pydocstyle D400.

================
Elastic Viewsets
================

.. autoclass:: resolwe.elastic.viewsets.ElasticSearchMixin
    :members:

"""
from types import MethodType

from elasticsearch_dsl.query import Q

from django.db.models import Case, IntegerField, Value, When

from guardian.utils import get_anonymous_user
from rest_framework.exceptions import APIException, ParseError
from rest_framework.response import Response
from rest_framework.viewsets import GenericViewSet

from .composer import composer
from .lookup import QueryBuilder
from .pagination import LimitOffsetPostPagination

__all__ = (
    'ElasticSearchMixin',
    'PaginationMixin',
    'ElasticSearchBaseViewSet',
)

ELASTICSEARCH_SIZE = 10000  # maximum number of results returned by ElasticSearch


class TooManyResults(APIException):
    """Exception when elastic query returns more than ``ELASTICSEARCH_SIZE`` results."""

    status_code = 400
    default_detail = 'Query returned too many results. Please, add more filters or use pagination.'
    default_code = 'bad_request'


class ElasticSearchMixin:
    """Mixin to use Django REST Framework with ElasticSearch based querysets.

    This mixin adds following methods:
      * :func:`~ElasticSearchMixin.order_search`
      * :func:`~ElasticSearchMixin.filter_search`
      * :func:`~ElasticSearchMixin.filter_permissions`

    """

    filtering_fields = []
    filtering_map = {}
    ordering_fields = []
    ordering_map = {}
    ordering = None

    def __init__(self, *args, **kwargs):
        """Construct viewset."""
        # Add registered viewset extensions. We take care not to modify the original
        # class-derived attributes.
        self.ordering_map = self.ordering_map.copy()

        for extension in composer.get_extensions(self):
            filtering_fields = getattr(extension, 'filtering_fields', [])
            filtering_map = getattr(extension, 'filtering_map', {})
            ordering_fields = getattr(extension, 'ordering_fields', [])
            ordering_map = getattr(extension, 'ordering_map', {})

            self.filtering_fields = self.filtering_fields + tuple(filtering_fields)
            self.filtering_map.update(filtering_map)
            self.ordering_fields = self.ordering_fields + tuple(ordering_fields)
            self.ordering_map.update(ordering_map)

            for field_name in filtering_fields:
                custom_filter_name = 'custom_filter_{}'.format(field_name)
                custom_filter = getattr(extension, custom_filter_name, None)
                if custom_filter:
                    setattr(self, custom_filter_name, MethodType(custom_filter, self))

        super().__init__(*args, **kwargs)

    def get_query_param(self, key, default=None):
        """Get query parameter uniformly for GET and POST requests."""
        value = self.request.query_params.get(key, None)
        if value is None:
            value = self.request.data.get(key, None)
        if value is None:
            value = default
        return value

    def get_query_params(self):
        """Get combined query parameters (GET and POST)."""
        params = self.request.query_params.copy()
        params.update(self.request.data)
        return params

    def order_search(self, search):
        """Order given search by the ordering parameter given in request.

        :param search: ElasticSearch query object

        """
        ordering = self.get_query_param('ordering', self.ordering)
        if not ordering:
            return search

        sort_fields = []
        for raw_ordering in ordering.split(','):
            ordering_field = raw_ordering.lstrip('-')
            if ordering_field not in self.ordering_fields:
                raise ParseError('Ordering by `{}` is not supported.'.format(ordering_field))

            ordering_field = self.ordering_map.get(ordering_field, ordering_field)
            direction = '-' if raw_ordering[0] == '-' else ''
            sort_fields.append('{}{}'.format(direction, ordering_field))

        return search.sort(*sort_fields)

    def get_always_allowed_arguments(self):
        """Return query arguments which are always allowed."""
        return [
            'ordering',
            'offset',
            'limit',
            'format',
            'fields',
        ]

    def filter_search(self, search):
        """Filter given search by the filter parameter given in request.

        :param search: ElasticSearch query object

        """
        builder = QueryBuilder(
            self.filtering_fields,
            self.filtering_map,
            self
        )
        search, unmatched = builder.build(search, self.get_query_params())

        # Ensure that no unsupported arguments were used.
        for argument in self.get_always_allowed_arguments():
            unmatched.pop(argument, None)

        if unmatched:
            raise ParseError('Unsupported filter arguments: {}'.format(', '.join(unmatched.keys())))

        return search

    def filter_permissions(self, search):
        """Filter given query based on permissions of the user in the request.

        :param search: ElasticSearch query object

        """
        user = self.request.user
        if user.is_superuser:
            return search
        if user.is_anonymous:
            user = get_anonymous_user()

        filters = [Q('match', users_with_permissions=user.pk)]
        filters.extend([
            Q('match', groups_with_permissions=group.pk) for group in user.groups.all()
        ])
        filters.append(Q('match', public_permission=True))

        # `minimum_should_match` is set to 1 by default
        return search.query('bool', should=filters)


class PaginationMixin:
    """Mixin for making paginated response in case pagination parameters are provided."""

    def paginate_response(self, queryset, serializers_kwargs={}):
        """Optionally return paginated response.

        If pagination parameters are provided in the request, then paginated response
        is returned, otherwise response is not paginated.

        """
        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True, **serializers_kwargs)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True, **serializers_kwargs)
        return Response(serializer.data)


class ElasticSearchBaseViewSet(PaginationMixin, ElasticSearchMixin, GenericViewSet):
    """Base ViewSet for ElasticSearch based views.

    This ViewSet creates search based on ``document_class`` parameter, specified
    in subclass and:

      * filter it by query parameters
      * apply ordering
      * filter permissions
      * apply pagination

    .. IMPORTANT::

        Both ``POST`` and ``GET`` requests are supported.
    """

    document_class = None

    pagination_class = LimitOffsetPostPagination

    def custom_filter(self, search):
        """Perform custom search filtering.

        This method is intended to be overriden in subclasses.
        """
        return search

    def search(self):
        """Handle the search request."""
        search = self.document_class().search()  # pylint: disable=not-callable

        search = self.custom_filter(search)

        search = self.filter_search(search)
        search = self.order_search(search)
        search = self.filter_permissions(search)

        if search.count() > ELASTICSEARCH_SIZE:
            limit = self.paginator.get_limit(self.request)

            if not limit or limit > ELASTICSEARCH_SIZE:
                raise TooManyResults()

        search = search.extra(size=ELASTICSEARCH_SIZE)
        return search

    def list_with_post(self, request):
        """Endpoint handler."""
        search = self.search()
        return self.paginate_response(search)

    def list(self, request):
        """Endpoint handler."""
        return self.list_with_post(request)


class ElasticSearchCombinedViewSet(ElasticSearchBaseViewSet):
    """ViewSet, which combines database and ES queries.

    To use this viewset, you should mixin the viewset that handles
    the database part, for example:

    .. code:: python

        class MyCombinedViewSet(ElasticSearchCombinedViewSet, MyViewSet):
            # ...

    Whenever the ``is_search_request`` method returns ``True`` (by default
    it always does), the search part is used. Otherwise, only the database
    part is used. In both cases, search results are only used to get the
    primary keys to filter the database query.

    In order for this to work, your index must store the primary key
    in a field (by default it should be called ``id``). You can change
    this by setting ``primary_key_field``.
    """

    primary_key_field = 'id'

    def is_search_request(self):
        """Check if current request is a search request."""
        return True

    def list_with_post(self, request):
        """Endpoint handler."""
        if self.is_search_request():
            search = self.search()

            page = self.paginate_queryset(search)
            if page is None:
                items = search
            else:
                items = page

            try:
                primary_keys = []
                order_map_cases = []
                for order, item in enumerate(items):
                    pk = item[self.primary_key_field]
                    primary_keys.append(pk)
                    order_map_cases.append(When(pk=pk, then=Value(order)))

                queryset = self.get_queryset().filter(
                    pk__in=primary_keys
                ).order_by(
                    Case(*order_map_cases, output_field=IntegerField()).asc()
                )
            except KeyError:
                raise KeyError("Combined viewset requires that your index contains a field with "
                               "the primary key. By default this field is called 'id', but you "
                               "can change it by setting primary_key_field.")

            # Pagination must be handled differently.
            serializer = self.get_serializer(queryset, many=True)
            if page is not None:
                return self.get_paginated_response(serializer.data)

            return Response(serializer.data)
        else:
            queryset = self.filter_queryset(self.get_queryset())
            return self.paginate_response(queryset)

    def list(self, request):
        """Endpoint handler."""
        return self.list_with_post(request)



# pylint: disable=missing-docstring
import mock

from django.contrib.auth.models import AnonymousUser
from django.contrib.contenttypes.models import ContentType
from django.core.urlresolvers import reverse
from django.db import DEFAULT_DB_ALIAS, connections
from django.test.utils import CaptureQueriesContext

from guardian.conf.settings import ANONYMOUS_USER_NAME
from guardian.models import UserObjectPermission
from guardian.shortcuts import assign_perm, remove_perm
from rest_framework import exceptions, status
from rest_framework.test import APIRequestFactory, force_authenticate

from resolwe.flow.models import Collection, Data, DescriptorSchema, Entity, Process
from resolwe.flow.views import CollectionViewSet, DataViewSet, EntityViewSet, ProcessViewSet
from resolwe.test import ResolweAPITestCase, TestCase

factory = APIRequestFactory()  # pylint: disable=invalid-name


MESSAGES = {
    'NOT_FOUND': 'Not found.',
    'NO_PERMS': 'You do not have permission to perform this action.',
}


class TestDataViewSetCase(TestCase):
    def setUp(self):
        super().setUp()

        self.data_viewset = DataViewSet.as_view(actions={
            'get': 'list',
            'post': 'create',
        })
        self.data_detail_viewset = DataViewSet.as_view(actions={
            'get': 'retrieve',
        })

        self.collection = Collection.objects.create(contributor=self.contributor)

        self.proc = Process.objects.create(
            type='data:test:process',
            slug='test-process',
            version='1.0.0',
            contributor=self.contributor,
            entity_type='test-schema',
            entity_descriptor_schema='test-schema',
            input_schema=[{'name': 'input_data', 'type': 'data:test:', 'required': False}],
        )

        self.descriptor_schema = DescriptorSchema.objects.create(
            slug='test-schema',
            version='1.0.0',
            contributor=self.contributor,
        )

        assign_perm('view_collection', self.contributor, self.collection)
        assign_perm('add_collection', self.contributor, self.collection)
        assign_perm('view_process', self.contributor, self.proc)
        assign_perm('view_descriptorschema', self.contributor, self.descriptor_schema)

    def test_prefetch(self):
        request = factory.get('/', '', format='json')
        force_authenticate(request, self.contributor)

        for _ in range(10):
            Data.objects.create(contributor=self.contributor, process=self.proc)

        # Check prefetch. The number of queries without prefetch depends
        # on the number of Data objects. With prefetch 56 queries,
        # without prefetch 73 queries. Python 2 and 3 have slightly
        # different number of queries, so we set a loose constraint in test.
        conn = connections[DEFAULT_DB_ALIAS]
        with CaptureQueriesContext(conn) as captured_queries:
            self.data_viewset(request)
            self.assertLess(len(captured_queries), 62)

    def test_descriptor_schema(self):
        # Descriptor schema can be assigned by slug.
        data = {'process': 'test-process', 'descriptor_schema': 'test-schema'}
        request = factory.post('/', data, format='json')
        force_authenticate(request, self.contributor)
        self.data_viewset(request)

        data = Data.objects.latest()
        self.assertEqual(data.descriptor_schema, self.descriptor_schema)

        # Descriptor schema can be assigned by id.
        data = {'process': 'test-process', 'descriptor_schema': self.descriptor_schema.pk}
        request = factory.post('/', data, format='json')
        force_authenticate(request, self.contributor)
        self.data_viewset(request)

        data = Data.objects.latest()
        self.assertEqual(data.descriptor_schema, self.descriptor_schema)

    def test_use_latest_with_perm(self):
        Process.objects.create(
            type='test:process',
            name='Test process',
            slug='test-process',
            version='2.0.0',
            contributor=self.contributor,
        )
        DescriptorSchema.objects.create(
            name='Test schema',
            slug='test-schema',
            version='2.0.0',
            contributor=self.contributor,
        )

        data = {'process': 'test-process', 'descriptor_schema': 'test-schema'}
        request = factory.post('/', data, format='json')
        force_authenticate(request, self.contributor)
        self.data_viewset(request)

        data = Data.objects.latest()
        # Check that older versions are user if user doesn't have permissions on the latest
        self.assertEqual(data.process, self.proc)
        self.assertEqual(data.descriptor_schema, self.descriptor_schema)

    def test_public_create(self):
        assign_perm('view_process', AnonymousUser(), self.proc)

        data = {'process': 'test-process'}
        request = factory.post('/', data, format='json')
        resp = self.data_viewset(request)
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)
        self.assertEqual(Data.objects.count(), 1)

        data = Data.objects.latest()
        self.assertEqual(data.contributor.username, ANONYMOUS_USER_NAME)
        self.assertEqual(data.process.slug, 'test-process')

    def test_inherit_permissions(self):
        data_ctype = ContentType.objects.get_for_model(Data)
        entity_ctype = ContentType.objects.get_for_model(Entity)

        assign_perm('view_collection', self.user, self.collection)
        assign_perm('add_collection', self.user, self.collection)

        post_data = {'process': 'test-process', 'collections': [self.collection.pk]}
        request = factory.post('/', post_data, format='json')
        force_authenticate(request, self.contributor)
        resp = self.data_viewset(request)
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)

        data = Data.objects.last()
        entity = Entity.objects.last()

        self.assertTrue(self.user.has_perm('view_data', data))
        self.assertTrue(self.user.has_perm('view_entity', entity))
        self.assertTrue(self.user.has_perm('add_entity', entity))
        self.assertEqual(UserObjectPermission.objects.filter(content_type=data_ctype, user=self.user).count(), 1)
        self.assertEqual(UserObjectPermission.objects.filter(content_type=entity_ctype, user=self.user).count(), 2)

        # Add some permissions and run another process in same entity.
        assign_perm('edit_collection', self.user, self.collection)
        assign_perm('share_entity', self.user, entity)

        post_data = {
            'process': 'test-process',
            'collections': [self.collection.pk],
            'input': {'input_data': data.pk},
        }
        request = factory.post('/', post_data, format='json')
        force_authenticate(request, self.contributor)
        resp = self.data_viewset(request)
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)

        data_2 = Data.objects.last()
        self.assertTrue(self.user.has_perm('view_data', data_2))
        self.assertTrue(self.user.has_perm('edit_data', data_2))
        self.assertTrue(self.user.has_perm('share_data', data_2))
        self.assertFalse(self.user.has_perm('edit_entity', entity))
        self.assertEqual(UserObjectPermission.objects.filter(content_type=data_ctype, user=self.user).count(), 4)
        self.assertEqual(UserObjectPermission.objects.filter(content_type=entity_ctype, user=self.user).count(), 3)

    def test_create_entity(self):
        data = {'process': 'test-process', 'collections': [self.collection.pk]}
        request = factory.post('/', data, format='json')
        force_authenticate(request, self.contributor)
        resp = self.data_viewset(request)
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)

        # Test that one Entity was created and that it was added to the same collection as Data object.
        self.assertEqual(Entity.objects.count(), 1)
        self.assertEqual(Entity.objects.first().collections.count(), 1)
        self.assertEqual(Entity.objects.first().collections.first().pk, self.collection.pk)

    def test_collections_fields(self):
        # Create data object.
        data = {'process': 'test-process', 'collections': [self.collection.pk]}
        request = factory.post('/', data, format='json')
        force_authenticate(request, self.contributor)
        response = self.data_viewset(request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = Data.objects.last()
        entity = Entity.objects.last()

        # Ensure collections/entities are not present in lists.
        request = factory.get('/', '', format='json')
        force_authenticate(request, self.contributor)
        response = self.data_viewset(request)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data), 1)
        self.assertNotIn('collections', response.data[0].keys())
        self.assertNotIn('entities', response.data[0].keys())

        # Check that query returns the correct collection ids.
        request = factory.get('/', '', format='json')
        force_authenticate(request, self.contributor)
        response = self.data_detail_viewset(request, pk=data.pk)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['collections'], [self.collection.pk])
        self.assertEqual(response.data['entities'], [entity.pk])

        # Check that hydrate_{collections,entities} works. Also ensure that the serializer
        # doesn't crash if hydrate_data is also set (could cause infinite recursion).
        request = factory.get('/', {
            'hydrate_collections': '1',
            'hydrate_entities': '1',
            'hydrate_data': '1',
        }, format='json')
        force_authenticate(request, self.contributor)
        response = self.data_detail_viewset(request, pk=data.pk)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['collections'][0]['id'], self.collection.pk)
        self.assertEqual(response.data['entities'][0]['id'], entity.pk)

    def test_process_is_active(self):
        # Do not allow creating data of inactive processes
        Process.objects.filter(slug='test-process').update(is_active=False)
        data = {'process': 'test-process'}
        request = factory.post('/', data, format='json')
        force_authenticate(request, self.contributor)
        response = self.data_viewset(request)
        self.assertEqual(response.status_code, 400)


class TestCollectionViewSetCase(TestCase):
    def setUp(self):
        super().setUp()

        self.checkslug_viewset = CollectionViewSet.as_view(actions={
            'get': 'slug_exists',
        })
        self.add_data_viewset = CollectionViewSet.as_view(actions={
            'post': 'add_data',
        })
        self.remove_data_viewset = CollectionViewSet.as_view(actions={
            'post': 'remove_data',
        })
        self.collection_detail_viewset = CollectionViewSet.as_view(actions={
            'get': 'retrieve',
            'put': 'update',
            'patch': 'partial_update',
            'delete': 'destroy',
        })
        self.collection_list_viewset = CollectionViewSet.as_view(actions={
            'get': 'list',
            'post': 'create',
        })

        self.detail_url = lambda pk: reverse('resolwe-api:collection-detail', kwargs={'pk': pk})

    def _create_data(self):
        process = Process.objects.create(
            name='Test process',
            contributor=self.contributor,
        )

        return Data.objects.create(
            name='Test data',
            contributor=self.contributor,
            process=process,
        )

    def _create_entity(self):
        return Entity.objects.create(
            name='Test entity',
            contributor=self.contributor,
        )

    def test_set_descriptor_schema(self):
        d_schema = DescriptorSchema.objects.create(slug="new-schema", name="New Schema", contributor=self.contributor)

        data = {
            'name': 'Test collection',
            'descriptor_schema': 'new-schema',
        }

        request = factory.post('/', data=data, format='json')
        force_authenticate(request, self.admin)
        self.collection_list_viewset(request)

        self.assertEqual(Collection.objects.count(), 1)
        self.assertEqual(Collection.objects.first().descriptor_schema, d_schema)

    def test_change_descriptor_schema(self):
        collection = Collection.objects.create(slug="collection1", name="Collection 1", contributor=self.contributor)
        d_schema = DescriptorSchema.objects.create(slug="new-schema", name="New Schema", contributor=self.contributor)

        request = factory.patch(self.detail_url(collection.pk), {'descriptor_schema': 'new-schema'}, format='json')
        force_authenticate(request, self.admin)
        self.collection_detail_viewset(request, pk=collection.pk)

        collection.refresh_from_db()
        self.assertEqual(collection.descriptor_schema, d_schema)

    def test_change_slug(self):
        collection1 = Collection.objects.create(name="Collection", contributor=self.contributor)
        collection2 = Collection.objects.create(name="Collection", contributor=self.contributor)
        self.assertEqual(collection1.slug, 'collection')
        self.assertEqual(collection2.slug, 'collection-2')

        request = factory.patch(self.detail_url(collection1.pk), {'name': 'Collection', 'slug': None}, format='json')
        force_authenticate(request, self.admin)
        response = self.collection_detail_viewset(request, pk=collection1.pk)
        self.assertEqual(response.data['slug'], 'collection')

        request = factory.patch(self.detail_url(collection2.pk), {'slug': 'collection-3'}, format='json')
        force_authenticate(request, self.admin)
        response = self.collection_detail_viewset(request, pk=collection2.pk)
        self.assertEqual(response.data['slug'], 'collection-3')

        request = factory.patch(self.detail_url(collection2.pk), {'slug': 'collection'}, format='json')
        force_authenticate(request, self.admin)
        response = self.collection_detail_viewset(request, pk=collection2.pk)
        self.assertContains(response, 'already taken', status_code=400)

    def test_check_slug(self):
        Collection.objects.create(slug="collection1", name="Collection 1", contributor=self.admin)

        # unauthorized
        request = factory.get('/', {'name': 'collection1'}, format='json')
        resp = self.checkslug_viewset(request)
        self.assertEqual(resp.status_code, 401)
        self.assertEqual(resp.data, None)

        # existing slug
        request = factory.get('/', {'name': 'collection1'}, format='json')
        force_authenticate(request, self.admin)
        resp = self.checkslug_viewset(request)
        self.assertEqual(resp.data, True)

        # existing slug - iexact
        request = factory.get('/', {'name': 'Collection1'}, format='json')
        force_authenticate(request, self.admin)
        resp = self.checkslug_viewset(request)
        self.assertEqual(resp.data, True)

        # non-existing slug
        request = factory.get('/', {'name': 'new-collection'}, format='json')
        force_authenticate(request, self.admin)
        resp = self.checkslug_viewset(request)
        self.assertEqual(resp.data, False)

        # bad query parameter
        request = factory.get('/', {'bad': 'parameter'}, format='json')
        force_authenticate(request, self.admin)
        resp = self.checkslug_viewset(request)
        self.assertEqual(resp.status_code, 400)

    def test_add_remove_data(self):
        c = Collection.objects.create(slug="collection1", name="Collection 1", contributor=self.contributor)
        assign_perm('view_collection', self.contributor, c)
        assign_perm('edit_collection', self.contributor, c)
        assign_perm('share_collection', self.contributor, c)

        proc = Process.objects.create(type='test:process', name='Test process', contributor=self.contributor)
        d = Data.objects.create(contributor=self.contributor, slug='test1', process=proc)

        request = factory.post(self.detail_url(c.pk), {'ids': [str(d.pk)]}, format='json')
        force_authenticate(request, self.contributor)

        # user w/o permissions cannot add data
        resp = self.add_data_viewset(request, pk=c.pk)
        self.assertEqual(resp.data['detail'], MESSAGES['NO_PERMS'])
        self.assertEqual(c.data.count(), 0)

        assign_perm('add_collection', self.contributor, c)

        # user w/ permissions can add data
        resp = self.add_data_viewset(request, pk=c.pk)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(c.data.count(), 1)

        request = factory.post(self.detail_url(c.pk), {'ids': [str(d.pk)]}, format='json')
        force_authenticate(request, self.contributor)
        remove_perm('add_collection', self.contributor, c)

        # user w/o permissions cannot remove data
        resp = self.remove_data_viewset(request, pk=c.pk)
        self.assertEqual(resp.data['detail'], MESSAGES['NO_PERMS'])
        self.assertEqual(c.data.count(), 1)

        assign_perm('add_collection', self.contributor, c)

        # user w/ permissions can remove data
        resp = self.remove_data_viewset(request, pk=c.pk)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(c.data.count(), 0)

        request = factory.post(self.detail_url(c.pk), {'ids': ['42']}, format='json')
        force_authenticate(request, self.contributor)

        resp = self.remove_data_viewset(request, pk=c.pk)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(c.data.count(), 0)

    def test_delete(self):
        collection = Collection.objects.create(
            name="Test collection",
            contributor=self.contributor,
        )

        data_1, data_2 = self._create_data(), self._create_data()
        entity_1, entity_2 = self._create_entity(), self._create_entity()

        collection.data.add(data_1, data_2)
        collection.entity_set.add(entity_1, entity_2)

        assign_perm("view_collection", self.user, collection)
        assign_perm("edit_collection", self.user, collection)
        assign_perm("view_data", self.user, data_1)
        assign_perm("view_data", self.user, data_2)
        assign_perm("edit_data", self.user, data_1)
        assign_perm("view_entity", self.user, entity_1)
        assign_perm("view_entity", self.user, entity_2)
        assign_perm("edit_entity", self.user, entity_1)

        request = factory.delete(self.detail_url(collection.pk))
        force_authenticate(request, self.user)
        self.collection_detail_viewset(request, pk=collection.pk)

        self.assertTrue(Data.objects.filter(pk=data_1.pk).exists())
        self.assertTrue(Data.objects.filter(pk=data_2.pk).exists())
        self.assertTrue(Entity.objects.filter(pk=entity_1.pk).exists())
        self.assertTrue(Entity.objects.filter(pk=entity_2.pk).exists())

        # Recreate the initial state and test with `delete_content` flag.
        collection = Collection.objects.create(
            name="Test collection",
            contributor=self.contributor,
        )

        collection.data.add(data_1, data_2)
        collection.entity_set.add(entity_1, entity_2)

        assign_perm("view_collection", self.user, collection)
        assign_perm("edit_collection", self.user, collection)

        request = factory.delete('{}?delete_content=1'.format(self.detail_url(collection.pk)))
        force_authenticate(request, self.user)
        self.collection_detail_viewset(request, pk=collection.pk)

        # Only objects with `edit` permission can be deleted.
        self.assertFalse(Data.objects.filter(pk=data_1.pk).exists())
        self.assertTrue(Data.objects.filter(pk=data_2.pk).exists())
        self.assertFalse(Entity.objects.filter(pk=entity_1.pk).exists())
        self.assertTrue(Entity.objects.filter(pk=entity_2.pk).exists())


class ProcessTestCase(ResolweAPITestCase):
    def setUp(self):
        self.resource_name = 'process'
        self.viewset = ProcessViewSet

        super().setUp()

    def test_create_new(self):
        post_data = {
            'name': 'Test process',
            'slug': 'test-process',
            'type': 'data:test:',
        }

        # Normal user is not allowed to create new processes.
        resp = self._post(post_data, self.contributor)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)

        # Superuser can create process.
        resp = self._post(post_data, self.admin)
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)

    def test_is_active(self):
        post_data = {
            'name': 'Test process',
            'slug': 'test-process',
            'type': 'data:test:',
            'is_active': False,
        }

        # is_active can not be set through API and is True by default
        response = self._post(post_data, self.admin)
        self.assertTrue(response.data['is_active'])

        # is_active should not be changed through API
        process_id = response.data['id']
        response = self._patch(process_id, {'is_active': False}, self.admin)
        self.assertEqual(response.status_code, 405)  # PATCH not allowed on process


class EntityViewSetTest(TestCase):
    def setUp(self):
        super().setUp()

        self.collection = Collection.objects.create(name="Test Collection", contributor=self.contributor)
        self.collection2 = Collection.objects.create(name="Test Collection 2", contributor=self.contributor)
        self.entity = Entity.objects.create(name="Test entity", contributor=self.contributor)
        process = Process.objects.create(name="Test process", contributor=self.contributor)
        self.data = Data.objects.create(name="Test data", contributor=self.contributor, process=process)
        self.data_2 = Data.objects.create(name="Test data 2", contributor=self.contributor, process=process)

        # another Data object to make sure that other objects are not processed
        Data.objects.create(name="Dummy data", contributor=self.contributor, process=process)

        self.entity.data.add(self.data)
        self.entity.collections.add(self.collection2)

        assign_perm('add_collection', self.contributor, self.collection)
        assign_perm('add_entity', self.contributor, self.entity)
        assign_perm('view_collection', self.contributor, self.collection)
        assign_perm('view_collection', self.contributor, self.collection2)
        assign_perm('view_entity', self.contributor, self.entity)

        self.entityviewset = EntityViewSet()

        self.entity_detail_viewset = EntityViewSet.as_view(actions={
            'get': 'retrieve',
            'put': 'update',
            'patch': 'partial_update',
            'delete': 'destroy',
        })
        self.entity_list_viewset = EntityViewSet.as_view(actions={
            'get': 'list',
            'post': 'create',
        })

        self.detail_url = lambda pk: reverse('resolwe-api:entity-detail', kwargs={'pk': pk})

    def _create_data(self):
        process = Process.objects.create(
            name='Test process',
            contributor=self.contributor,
        )

        return Data.objects.create(
            name='Test data',
            contributor=self.contributor,
            process=process,
        )

    def test_list_filter_collections(self):
        request = factory.get('/', {}, format='json')
        force_authenticate(request, self.contributor)
        resp = self.entity_list_viewset(request)
        self.assertEqual(len(resp.data), 1)

        request = factory.get('/', {'collections': 999999}, format='json')
        force_authenticate(request, self.contributor)
        resp = self.entity_list_viewset(request)
        self.assertEqual(len(resp.data), 0)

        request = factory.get('/', {'collections': self.collection.pk}, format='json')
        force_authenticate(request, self.contributor)
        resp = self.entity_list_viewset(request)
        self.assertEqual(len(resp.data), 0)

        request = factory.get('/', {'collections': self.collection2.pk}, format='json')
        force_authenticate(request, self.contributor)
        resp = self.entity_list_viewset(request)
        self.assertEqual(len(resp.data), 1)

    def test_add_to_collection(self):
        request_mock = mock.MagicMock(data={'ids': [self.collection.pk]}, user=self.contributor)
        self.entityviewset.get_object = lambda: self.entity

        self.assertEqual(self.entity.collections.count(), 1)

        self.entityviewset.add_to_collection(request_mock)

        self.assertEqual(self.collection.data.count(), 1)
        self.assertEqual(self.entity.collections.count(), 2)

    def test_remove_from_collection(self):
        # Manually add Entity and it's Data objects to the Collection
        self.entity.collections.add(self.collection.pk)
        self.collection.data.add(self.data)

        request_mock = mock.MagicMock(data={'ids': [self.collection.pk]}, user=self.contributor)
        self.entityviewset.get_object = lambda: self.entity

        self.assertEqual(self.entity.collections.count(), 2)

        self.entityviewset.remove_from_collection(request_mock)

        self.assertEqual(self.collection.data.count(), 0)
        self.assertEqual(self.entity.collections.count(), 1)

    def test_add_remove_permissions(self):
        request_mock = mock.MagicMock(data={'ids': [self.collection.pk]}, user=self.contributor)
        self.entityviewset.get_object = lambda: self.entity

        remove_perm('add_collection', self.contributor, self.collection)

        with self.assertRaises(exceptions.PermissionDenied):
            self.entityviewset.remove_from_collection(request_mock)

        with self.assertRaises(exceptions.PermissionDenied):
            self.entityviewset.add_to_collection(request_mock)

    def test_add_data(self):
        self.entity.collections.add(self.collection)

        request_mock = mock.MagicMock(data={'ids': [self.data.pk]}, user=self.contributor)
        self.entityviewset.get_object = lambda: self.entity

        self.entityviewset.add_data(request_mock)

        self.assertEqual(self.entity.data.count(), 1)
        self.assertEqual(self.collection.data.count(), 1)

    def test_delete(self):
        entity = Entity.objects.create(
            name="Test entity",
            contributor=self.contributor,
        )

        data_1, data_2 = self._create_data(), self._create_data()

        entity.data.add(data_1, data_2)

        assign_perm('view_entity', self.user, entity)
        assign_perm('edit_entity', self.user, entity)
        assign_perm('view_data', self.user, data_1)
        assign_perm('view_data', self.user, data_2)
        assign_perm('edit_data', self.user, data_1)

        request = factory.delete(self.detail_url(entity.pk))
        force_authenticate(request, self.user)
        self.entity_detail_viewset(request, pk=entity.pk)

        self.assertTrue(Data.objects.filter(pk=data_1.pk).exists())
        self.assertTrue(Data.objects.filter(pk=data_2.pk).exists())

        # Recreate the initial state and test with `delete_content` flag.
        entity = Entity.objects.create(
            name="Test entity",
            contributor=self.contributor,
        )

        entity.data.add(data_1, data_2)

        assign_perm('view_entity', self.user, entity)
        assign_perm('edit_entity', self.user, entity)

        request = factory.delete('{}?delete_content=1'.format(self.detail_url(entity.pk)))
        force_authenticate(request, self.user)
        self.entity_detail_viewset(request, pk=entity.pk)

        # Only objects with `edit` permission can be deleted.
        self.assertFalse(Data.objects.filter(pk=data_1.pk).exists())
        self.assertTrue(Data.objects.filter(pk=data_2.pk).exists())

        # Ensure that deletion works correctly when all data objects of an entity
        # are deleted.
        entity = Entity.objects.create(
            name="Test entity",
            contributor=self.contributor,
        )

        assign_perm('view_entity', self.user, entity)
        assign_perm('edit_entity', self.user, entity)
        assign_perm('edit_data', self.user, data_2)

        entity.data.add(data_2)

        request = factory.delete('{}?delete_content=1'.format(self.detail_url(entity.pk)))
        force_authenticate(request, self.user)
        response = self.entity_detail_viewset(request, pk=entity.pk)

        self.assertEqual(response.status_code, 204)
        self.assertFalse(Entity.objects.filter(pk=entity.pk).exists())
        self.assertFalse(Data.objects.filter(pk=data_2.pk).exists())



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-09-05 09:28
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0017_refactor_relations_3'),
    ]

    operations = [
        migrations.AddField(
            model_name='process',
            name='is_active',
            field=models.BooleanField(default=True, verbose_name='active'),
        ),
    ]



"""Resolwe collection serializer."""
from django.contrib.auth.models import AnonymousUser

from rest_framework import serializers

from resolwe.flow.models import Collection
from resolwe.permissions.shortcuts import get_objects_for_user
from resolwe.rest.fields import ProjectableJSONField

from .base import ResolweBaseSerializer
from .data import DataSerializer
from .fields import NestedDescriptorSchemaSerializer


class CollectionSerializer(ResolweBaseSerializer):
    """Serializer for Collection objects."""

    settings = ProjectableJSONField(required=False)
    descriptor = ProjectableJSONField(required=False)
    descriptor_schema = NestedDescriptorSchemaSerializer(required=False)

    class Meta:
        """CollectionSerializer Meta options."""

        model = Collection
        read_only_fields = (
            'created',
            'descriptor_dirty',
            'id',
            'modified',
        )
        update_protected_fields = (
            'contributor',
        )
        fields = read_only_fields + update_protected_fields + (
            'data',
            'description',
            'descriptor',
            'descriptor_schema',
            'name',
            'settings',
            'slug',
            'tags',
        )

    def _serialize_data(self, data):
        """Return serialized data or list of ids, depending on `hydrate_data` query param."""
        if self.request and self.request.query_params.get('hydrate_data', False):
            serializer = DataSerializer(data, many=True, read_only=True)
            serializer.bind('data', self)
            return serializer.data
        else:
            return [d.id for d in data]

    def _filter_queryset(self, perms, queryset):
        """Filter object objects by permissions of user in request."""
        user = self.request.user if self.request else AnonymousUser()
        return get_objects_for_user(user, perms, queryset)

    def get_data(self, collection):
        """Return serialized list of data objects on collection that user has `view` permission on."""
        data = self._filter_queryset('view_data', collection.data.all())

        return self._serialize_data(data)

    def get_fields(self):
        """Dynamically adapt fields based on the current request."""
        fields = super(CollectionSerializer, self).get_fields()

        if self.request.method == "GET":
            fields['data'] = serializers.SerializerMethodField()
        else:
            fields['data'] = serializers.PrimaryKeyRelatedField(many=True, read_only=True)

        return fields



# pylint: disable=missing-docstring
import os
import subprocess
import unittest

import mock

from django.conf import settings
from django.db import transaction
from django.test import override_settings

from guardian.shortcuts import assign_perm

from resolwe.flow.executors.prepare import BaseFlowExecutorPreparer
from resolwe.flow.managers import manager
from resolwe.flow.models import Collection, Data, DataDependency, Process
from resolwe.test import ProcessTestCase, TestCase, tag_process, with_docker_executor, with_null_executor

# Workaround for false positive warnings in pylint.
# TODO: Can be removed with Django 2.0 or when pylint issue is fixed:
#       https://github.com/PyCQA/pylint/issues/1653
# pylint: disable=deprecated-method

PROCESSES_DIR = os.path.join(os.path.dirname(__file__), 'processes')
DESCRIPTORS_DIR = os.path.join(os.path.dirname(__file__), 'descriptors')


class GetToolsTestCase(TestCase):

    @mock.patch('resolwe.flow.utils.apps')
    @mock.patch('resolwe.flow.utils.os')
    @mock.patch('resolwe.flow.utils.settings')
    def test_get_tools_paths(self, settings_mock, os_mock, apps_mock):
        apps_mock.get_app_configs.return_value = [
            mock.MagicMock(path='/resolwe/test_app1'),
            mock.MagicMock(path='/resolwe/test_app2'),
        ]
        os_mock.path.join = os.path.join
        os_mock.path.isdir.side_effect = [False, True]
        settings_mock.RESOLWE_CUSTOM_TOOLS_PATHS = ['/custom_tools']

        base_executor = BaseFlowExecutorPreparer()
        tools_list = base_executor.get_tools_paths()

        self.assertEqual(len(tools_list), 2)
        self.assertIn('/resolwe/test_app2/tools', tools_list)
        self.assertIn('/custom_tools', tools_list)

    @mock.patch('resolwe.flow.utils.apps')
    @mock.patch('resolwe.flow.utils.settings')
    def test_not_list(self, settings_mock, apps_mock):
        apps_mock.get_app_configs.return_value = []
        settings_mock.RESOLWE_CUSTOM_TOOLS_PATHS = '/custom_tools'

        base_executor = BaseFlowExecutorPreparer()
        with self.assertRaisesRegex(KeyError, 'setting must be a list'):
            base_executor.get_tools_paths()


class ManagerRunProcessTest(ProcessTestCase):
    def setUp(self):
        super().setUp()
        self._register_schemas(processes_paths=[PROCESSES_DIR], descriptors_paths=[DESCRIPTORS_DIR])

    @tag_process('test-min')
    def test_minimal_process(self):
        self.run_process('test-min')

    @tag_process('test-missing-file')
    def test_missing_file(self):
        self.run_process('test-missing-file', assert_status=Data.STATUS_ERROR)

        data = Data.objects.last()
        self.assertEqual(data.status, Data.STATUS_ERROR)
        self.assertEqual(len(data.process_error), 1)
        self.assertIn('Referenced file does not exist', data.process_error[0])

    @tag_process('test-spawn-new')
    def test_spawn(self):
        self.run_process('test-spawn-new', tags=['test-tag'])

        data = Data.objects.last()
        data_dir = settings.FLOW_EXECUTOR['DATA_DIR']
        file_path = os.path.join(data_dir, str(data.pk), 'foo.bar')
        self.assertEqual(data.output['saved_file']['file'], 'foo.bar')
        self.assertTrue(os.path.isfile(file_path))
        self.assertEqual(data.tags, ['test-tag'])

        parent_data = Data.objects.first()
        self.assertEqual(data.parents.count(), 1)
        self.assertEqual(data.parents.first(), parent_data)

        # Check correct dependency type is created.
        self.assertEqual({d.kind for d in data.parents_dependency.all()}, {DataDependency.KIND_SUBPROCESS})

    @unittest.skipIf(True, "since PR308: the exception happens in other processes, currently impossible to propagate")
    @tag_process('test-spawn-missing-file')
    def test_spawn_missing_export(self):
        with self.assertRaisesRegex(KeyError, 'Use `re-export`'):
            self.run_process('test-spawn-missing-file')

    @tag_process('test-broken', 'test-broken-invalid-execution-engine', 'test-broken-invalid-expression-engine',
                 'test-broken-data-name')
    def test_broken(self):
        Process.objects.create(
            slug='test-broken-invalid-execution-engine',
            name='Test Process',
            contributor=self.contributor,
            type='data:test',
            version=1,
            run={
                'language': 'invalid',
            }
        )

        self.run_process('test-broken', assert_status=Data.STATUS_ERROR)
        self.run_process('test-broken-invalid-expression-engine', assert_status=Data.STATUS_ERROR)
        self.run_process('test-broken-invalid-execution-engine', assert_status=Data.STATUS_ERROR)

        # If evaluation of data_name template fails, the process should not abort as the
        # template may be evaluatable later when the process completes.
        self.run_process('test-broken-data-name')

    @tag_process('test-broken-invalide-storage')
    def test_invalid_storage_file(self):
        data = self.run_process('test-broken-invalide-storage', assert_status=Data.STATUS_ERROR)

        self.assertEqual(data.status, Data.STATUS_ERROR)
        self.assertIn("Value of 'storage' must be a valid JSON, current: 1a", data.process_error)

    @tag_process('test-workflow-1')
    def test_workflow(self):
        with transaction.atomic():
            # We need this transaction to delay calling the manager until we've
            # assigned permissions
            self.run_process('test-workflow-1', {'param1': 'world'}, tags=['test-tag'])
            workflow_data = Data.objects.get(process__slug='test-workflow-1')

            assign_perm('view_data', self.contributor, workflow_data)
            assign_perm('view_data', self.group, workflow_data)

        workflow_data.refresh_from_db()
        step1_data = Data.objects.get(process__slug='test-example-1')
        step2_data = Data.objects.get(process__slug='test-example-2')

        # Workflow should output indices of all data objects, in order.
        self.assertEqual(workflow_data.output['steps'], [step1_data.pk, step2_data.pk])

        # Steps should execute with the correct variables.
        self.assertEqual(step1_data.input['param1'], 'world')
        self.assertEqual(step1_data.input['param2'], True)
        self.assertEqual(step1_data.output['out1'], 'hello world')
        self.assertEqual(step1_data.tags, ['test-tag'])
        self.assertEqual(step2_data.input['param1'], step1_data.pk)
        self.assertEqual(step2_data.input['param2']['a'], step1_data.pk)
        self.assertEqual(step2_data.input['param2']['b'], 'hello')
        self.assertEqual(step2_data.output['out1'], 'simon says: hello world')
        self.assertEqual(step2_data.tags, ['test-tag'])

        self.assertEqual(step1_data.parents.count(), 1)
        self.assertEqual(step1_data.parents.first(), workflow_data)
        self.assertEqual(step2_data.parents.count(), 2)
        self.assertCountEqual(step2_data.parents.all(), [workflow_data, step1_data])

        # Check correct dependency type is created.
        self.assertEqual({d.kind for d in step1_data.parents_dependency.all()}, {DataDependency.KIND_SUBPROCESS})
        self.assertEqual({d.kind for d in step2_data.parents_dependency.all()},
                         {DataDependency.KIND_SUBPROCESS, DataDependency.KIND_IO})

        self.assertTrue(self.contributor.has_perm('flow.view_data', step1_data))
        # User inherites permission from group
        self.assertTrue(self.user.has_perm('flow.view_data', step1_data))

    @tag_process('test-workflow-2')
    def test_workflow_entity(self):
        collection = Collection.objects.create(contributor=self.contributor)

        with transaction.atomic():
            # We need this transaction to delay calling the manager until we've
            # add data to the collection.
            with self.preparation_stage():
                input_data = self.run_process('test-example-3')

            collection.data.add(input_data)
            input_data.entity_set.first().collections.add(collection)

        workflow_data = self.run_process('test-workflow-2', {'data1': input_data.pk})

        # Check that workflow results are added to the collection.
        self.assertEqual(collection.data.filter(pk__in=workflow_data.output['steps']).count(), 1)
        self.assertEqual(collection.data.all().count(), 2)

    @with_docker_executor
    @tag_process('test-docker')
    def test_run_in_docker(self):
        data = self.run_process('test-docker')
        self.assertEqual(data.output['result'], 'OK')

    @with_docker_executor
    @tag_process('test-requirements-docker')
    def test_executor_requirements(self):
        data = self.run_process('test-requirements-docker')
        self.assertEqual(data.output['result'], 'OK')

    @with_docker_executor
    @tag_process('test-docker-uid-gid')
    def test_docker_uid_gid(self):
        data = self.run_process('test-docker-uid-gid')
        self.assertEqual(data.output['result'], 'OK')

    @with_null_executor
    @tag_process('test-save-number')
    def test_null_executor(self):
        data = self.run_process('test-save-number', {'number': 19}, assert_status=Data.STATUS_WAITING)
        self.assertEqual(data.input['number'], 19)
        self.assertEqual(data.output, {})

    # TODO: Debug why the 'test-memory-resource-alloc' process doesn't end with and error on Travis
    @unittest.skipIf(os.environ.get('TRAVIS', '') == 'true', "Fails on Travis CI")
    @with_docker_executor
    @tag_process('test-memory-resource-alloc', 'test-memory-resource-noalloc')
    def test_memory_resource(self):
        # This process should be terminated due to too much memory usage.
        self.run_process('test-memory-resource-alloc', assert_status=Data.STATUS_ERROR)
        # This process should run normally (honors the limits).
        self.run_process('test-memory-resource-noalloc')

    @with_docker_executor
    @tag_process('test-cpu-resource-1core', 'test-cpu-resource-2core')
    def test_cpu_resource(self):
        # Currently there is no good way to test this reliably, so we just check if the
        # resource limit specification still makes the process run.
        self.run_process('test-cpu-resource-1core')
        self.run_process('test-cpu-resource-2core')

    @unittest.skipIf(settings.FLOW_DOCKER_DISABLE_SECCOMP, "Docker seccomp is disabled.")
    @with_docker_executor
    @tag_process('test-network-resource-enabled', 'test-network-resource-disabled', 'test-network-resource-policy')
    def test_network_resource(self):
        self.run_process('test-network-resource-enabled')
        self.run_process('test-network-resource-disabled', assert_status=Data.STATUS_ERROR)
        self.run_process('test-network-resource-policy', assert_status=Data.STATUS_ERROR)

    @with_docker_executor
    @override_settings(FLOW_PROCESS_RESOURCE_DEFAULTS={'cpu_time_interactive': 1})
    @tag_process('test-scheduling-class-interactive-ok', 'test-scheduling-class-interactive-fail',
                 'test-scheduling-class-batch')
    def test_scheduling_class(self):
        self.run_process('test-scheduling-class-interactive-ok')
        self.run_process('test-scheduling-class-interactive-fail', assert_status=Data.STATUS_ERROR)
        self.run_process('test-scheduling-class-batch')

    @with_docker_executor
    @tag_process('test-save-number')
    def test_executor_fs_lock(self):
        # First, run the process normaly.
        data = self.run_process('test-save-number', {'number': 42})

        # Make sure that process was successfully ran first time.
        self.assertEqual(data.output['number'], 42)
        data.output = {}
        data.save()

        process = subprocess.run(
            ['python', '-m', 'executors', '.docker'],
            cwd=os.path.join(settings.FLOW_EXECUTOR['RUNTIME_DIR'], str(data.pk)),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            timeout=5,
        )

        # Drain control events generated by this executor.
        manager.reset(keep_state=True)

        self.assertEqual(process.returncode, 0)

        # Check the status of the data object.
        data.refresh_from_db()
        # Check that output is empty and thus process didn't ran.
        self.assertEqual(data.output, {})
        self.assertEqual(data.status, Data.STATUS_DONE)
        self.assertEqual(data.process_error, [])



"""Elastic Search base index types for Resolwe models."""
import elasticsearch_dsl as dsl

from resolwe.elastic import indices
from resolwe.elastic.fields import Name, Slug, User
from resolwe.permissions.shortcuts import get_users_with_permission
from resolwe.permissions.utils import get_full_perm


class BaseDocument(indices.BaseDocument):
    """Base search document."""

    id = dsl.Integer()  # pylint: disable=invalid-name
    slug = Slug()
    version = dsl.Keyword()
    name = Name()
    created = dsl.Date()
    modified = dsl.Date()
    contributor_id = dsl.Integer()
    contributor_name = User()
    # We use a separate field for contributor sorting because we use an entirely
    # different value for it (the display name).
    contributor_sort = dsl.Keyword()
    owner_ids = dsl.Integer(multi=True)
    owner_names = User(multi=True)


class BaseIndexMixin:
    """Base index for objects used in ``BaseDocument``."""

    def get_contributor_sort_value(self, obj):
        """Generate display name for contributor."""
        user = obj.contributor

        if user.first_name or user.last_name:
            contributor = user.get_full_name()
        else:
            contributor = user.username

        return contributor.strip().lower()

    def _get_user(self, user):
        """Generate user filtering tokens."""
        return ' '.join([user.username, user.first_name, user.last_name])

    def get_contributor_id_value(self, obj):
        """Extract contributor identifier."""
        return obj.contributor.pk

    def get_contributor_name_value(self, obj):
        """Extract contributor name."""
        return self._get_user(obj.contributor)

    def get_owner_ids_value(self, obj):
        """Extract owners' ids."""
        return [
            user.pk
            for user in get_users_with_permission(obj, get_full_perm('owner', obj))
        ]

    def get_owner_names_value(self, obj):
        """Extract owners' names."""
        return [
            self._get_user(user)
            for user in get_users_with_permission(obj, get_full_perm('owner', obj))
        ]

    def get_version_value(self, obj):
        """Extract version metadata."""
        return str(obj.version)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-09-21 09:56
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0002_set_process_owners'),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(state_operations=[
            migrations.CreateModel(
                name='DataDependency',
                fields=[
                    ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ],
                options={
                    'db_table': 'flow_data_parents',
                },
            ),
            migrations.AlterField(
                model_name='data',
                name='parents',
                field=models.ManyToManyField(related_name='children', through='flow.DataDependency', to='flow.Data'),
            ),
            migrations.AddField(
                model_name='datadependency',
                name='child',
                field=models.ForeignKey(db_column='from_data_id', on_delete=django.db.models.deletion.CASCADE, related_name='+', to='flow.Data'),
            ),
            migrations.AddField(
                model_name='datadependency',
                name='parent',
                field=models.ForeignKey(db_column='to_data_id', on_delete=django.db.models.deletion.CASCADE, related_name='+', to='flow.Data'),
            ),
        ])
    ]
